"""
Elasticsearch - A distributed RESTful search and analytics server

Module to provide Elasticsearch compatibility to Salt
(compatible with Elasticsearch version 1.5.2+)

.. versionadded:: 2015.8.0

:depends:       `elasticsearch-py <http://elasticsearch-py.readthedocs.org/en/latest/>`_

:configuration: This module accepts connection configuration details either as
    parameters or as configuration settings in /etc/salt/minion on the relevant
    minions:

    .. code-block:: yaml

        elasticsearch:
          host: '10.10.10.100:9200'

        elasticsearch-cluster:
          hosts:
            - '10.10.10.100:9200'
            - '10.10.10.101:9200'
            - '10.10.10.102:9200'

        elasticsearch-extra:
          hosts:
            - '10.10.10.100:9200'
          use_ssl: True
          verify_certs: True
          ca_certs: /path/to/custom_ca_bundle.pem
          number_of_shards: 1
          number_of_replicas: 0
          functions_blacklist:
            - 'saltutil.find_job'
            - 'pillar.items'
            - 'grains.items'
          proxies:
            - http: http://proxy:3128
            - https: http://proxy:1080

    When specifying proxies the requests backend will be used and the 'proxies'
    data structure is passed as-is to that module.

    This data can also be passed into pillar. Options passed into opts will
    overwrite options passed into pillar.

    Some functionality might be limited by elasticsearch-py and Elasticsearch server versions.
"""

import logging

from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

try:
    import elasticsearch

    # pylint: disable=no-name-in-module
    from elasticsearch import RequestsHttpConnection

    # pylint: enable=no-name-in-module

    logging.getLogger("elasticsearch").setLevel(logging.CRITICAL)
    HAS_ELASTICSEARCH = True
except ImportError:
    HAS_ELASTICSEARCH = False


def __virtual__():
    """
    Only load if elasticsearch libraries exist.
    """
    if not HAS_ELASTICSEARCH:
        return (
            False,
            "Cannot load module elasticsearch: elasticsearch libraries not found",
        )
    return True


def _get_instance(hosts=None, profile=None):
    """
    Return the elasticsearch instance
    """
    es = None
    proxies = None
    use_ssl = False
    ca_certs = None
    verify_certs = True
    http_auth = None
    timeout = 10

    if profile is None:
        profile = "elasticsearch"

    if isinstance(profile, str):
        _profile = __salt__["config.option"](profile, None)
    elif isinstance(profile, dict):
        _profile = profile
    if _profile:
        hosts = _profile.get("host", hosts)
        if not hosts:
            hosts = _profile.get("hosts", hosts)
        proxies = _profile.get("proxies", None)
        use_ssl = _profile.get("use_ssl", False)
        ca_certs = _profile.get("ca_certs", None)
        verify_certs = _profile.get("verify_certs", True)
        username = _profile.get("username", None)
        password = _profile.get("password", None)
        timeout = _profile.get("timeout", 10)

        if username and password:
            http_auth = (username, password)

    if not hosts:
        hosts = ["127.0.0.1:9200"]
    if isinstance(hosts, str):
        hosts = [hosts]
    try:
        if proxies:
            # Custom connection class to use requests module with proxies
            class ProxyConnection(RequestsHttpConnection):
                def __init__(self, *args, **kwargs):
                    proxies = kwargs.pop("proxies", {})
                    super().__init__(*args, **kwargs)
                    self.session.proxies = proxies

            es = elasticsearch.Elasticsearch(
                hosts,
                connection_class=ProxyConnection,
                proxies=proxies,
                use_ssl=use_ssl,
                ca_certs=ca_certs,
                verify_certs=verify_certs,
                http_auth=http_auth,
                timeout=timeout,
            )
        else:
            es = elasticsearch.Elasticsearch(
                hosts,
                use_ssl=use_ssl,
                ca_certs=ca_certs,
                verify_certs=verify_certs,
                http_auth=http_auth,
                timeout=timeout,
            )

        # Try the connection
        es.info()
    except elasticsearch.exceptions.TransportError as err:
        raise CommandExecutionError(
            "Could not connect to Elasticsearch host/ cluster {} due to {}".format(
                hosts, err
            )
        )
    return es


def ping(allow_failure=False, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Test connection to Elasticsearch instance. This method does not fail if not explicitly specified.

    allow_failure
        Throw exception if ping fails

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.ping allow_failure=True
        salt myminion elasticsearch.ping profile=elasticsearch-extra
    """
    try:
        _get_instance(hosts, profile)
    except CommandExecutionError as e:
        if allow_failure:
            raise
        return False
    return True


def info(hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Return Elasticsearch information.

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.info
        salt myminion elasticsearch.info profile=elasticsearch-extra
    """
    es = _get_instance(hosts, profile)

    try:
        return es.info()
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve server information, server returned code {} with"
            " message {}".format(e.status_code, e.error)
        )


def node_info(nodes=None, flat_settings=False, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Return Elasticsearch node information.

    nodes
        List of cluster nodes (id or name) to display stats for. Use _local for connected node, empty for all
    flat_settings
        Flatten settings keys

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.node_info flat_settings=True
    """
    es = _get_instance(hosts, profile)

    try:
        return es.nodes.info(node_id=nodes, flat_settings=flat_settings)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve node information, server returned code {} with message {}".format(
                e.status_code, e.error
            )
        )


def cluster_health(index=None, level="cluster", local=False, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Return Elasticsearch cluster health.

    index
        Limit the information returned to a specific index
    level
        Specify the level of detail for returned information, default 'cluster', valid choices are: 'cluster', 'indices', 'shards'
    local
        Return local information, do not retrieve the state from master node

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.cluster_health
    """
    es = _get_instance(hosts, profile)

    try:
        return es.cluster.health(index=index, level=level, local=local)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve health information, server returned code {} with"
            " message {}".format(e.status_code, e.error)
        )


def cluster_stats(nodes=None, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Return Elasticsearch cluster stats.

    nodes
        List of cluster nodes (id or name) to display stats for. Use _local for connected node, empty for all

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.cluster_stats
    """
    es = _get_instance(hosts, profile)

    try:
        return es.cluster.stats(node_id=nodes)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve cluster stats, server returned code {} with message {}".format(
                e.status_code, e.error
            )
        )


def cluster_get_settings(
    flat_settings=False, include_defaults=False, hosts=None, profile=None
):
    """
    .. versionadded:: 3000

    Return Elasticsearch cluster settings.

    flat_settings
        Return settings in flat format.

    include_defaults
        Whether to return all default clusters setting.

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.cluster_get_settings
    """
    es = _get_instance(hosts, profile)

    try:
        return es.cluster.get_settings(
            flat_settings=flat_settings, include_defaults=include_defaults
        )
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve cluster settings, server returned code {} with message {}".format(
                e.status_code, e.error
            )
        )


def cluster_put_settings(body=None, flat_settings=False, hosts=None, profile=None):
    """
    .. versionadded:: 3000

    Set Elasticsearch cluster settings.

    body
        The settings to be updated. Can be either 'transient' or 'persistent' (survives cluster restart)
        http://www.elastic.co/guide/en/elasticsearch/reference/current/cluster-update-settings.html

    flat_settings
        Return settings in flat format.

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.cluster_put_settings '{"persistent": {"indices.recovery.max_bytes_per_sec": "50mb"}}'
        salt myminion elasticsearch.cluster_put_settings '{"transient": {"indices.recovery.max_bytes_per_sec": "50mb"}}'
    """
    if not body:
        message = "You must provide a body with settings"
        raise SaltInvocationError(message)
    es = _get_instance(hosts, profile)

    try:
        return es.cluster.put_settings(body=body, flat_settings=flat_settings)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot update cluster settings, server returned code {} with message {}".format(
                e.status_code, e.error
            )
        )


def alias_create(indices, alias, hosts=None, body=None, profile=None, source=None):
    """
    Create an alias for a specific index/indices

    indices
        Single or multiple indices separated by comma, use _all to perform the operation on all indices.
    alias
        Alias name
    body
        Optional definition such as routing or filter as defined in https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-aliases.html
    source
        URL of file specifying optional definition such as routing or filter. Cannot be used in combination with ``body``.

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.alias_create testindex_v1 testindex
    """
    es = _get_instance(hosts, profile)
    if source and body:
        message = "Either body or source should be specified but not both."
        raise SaltInvocationError(message)
    if source:
        body = __salt__["cp.get_file_str"](
            source, saltenv=__opts__.get("saltenv", "base")
        )
    try:
        result = es.indices.put_alias(index=indices, name=alias, body=body)
        return result.get("acknowledged", False)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot create alias {} in index {}, server returned code {} with"
            " message {}".format(alias, indices, e.status_code, e.error)
        )


def alias_delete(indices, aliases, hosts=None, body=None, profile=None, source=None):
    """
    Delete an alias of an index

    indices
        Single or multiple indices separated by comma, use _all to perform the operation on all indices.
    aliases
        Alias names separated by comma

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.alias_delete testindex_v1 testindex
    """
    es = _get_instance(hosts, profile)
    if source and body:
        message = "Either body or source should be specified but not both."
        raise SaltInvocationError(message)
    if source:
        body = __salt__["cp.get_file_str"](
            source, saltenv=__opts__.get("saltenv", "base")
        )
    try:
        result = es.indices.delete_alias(index=indices, name=aliases)

        return result.get("acknowledged", False)
    except elasticsearch.exceptions.NotFoundError:
        return True
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot delete alias {} in index {}, server returned code {} with"
            " message {}".format(aliases, indices, e.status_code, e.error)
        )


def alias_exists(aliases, indices=None, hosts=None, profile=None):
    """
    Return a boolean indicating whether given alias exists

    indices
        Single or multiple indices separated by comma, use _all to perform the operation on all indices.
    aliases
        Alias names separated by comma

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.alias_exists None testindex
    """
    es = _get_instance(hosts, profile)
    try:
        return es.indices.exists_alias(name=aliases, index=indices)
    except elasticsearch.exceptions.NotFoundError:
        return False
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot get alias {} in index {}, server returned code {} with message {}".format(
                aliases, indices, e.status_code, e.error
            )
        )


def alias_get(indices=None, aliases=None, hosts=None, profile=None):
    """
    Check for the existence of an alias and if it exists, return it

    indices
        Single or multiple indices separated by comma, use _all to perform the operation on all indices.
    aliases
        Alias names separated by comma

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.alias_get testindex
    """
    es = _get_instance(hosts, profile)

    try:
        return es.indices.get_alias(index=indices, name=aliases)
    except elasticsearch.exceptions.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot get alias {} in index {}, server returned code {} with message {}".format(
                aliases, indices, e.status_code, e.error
            )
        )


def document_create(
    index, doc_type, body=None, id=None, hosts=None, profile=None, source=None
):
    """
    Create a document in a specified index

    index
        Index name where the document should reside
    doc_type
        Type of the document
    body
        Document to store
    source
        URL of file specifying document to store. Cannot be used in combination with ``body``.
    id
        Optional unique document identifier for specified doc_type (empty for random)

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.document_create testindex doctype1 '{}'
    """
    es = _get_instance(hosts, profile)
    if source and body:
        message = "Either body or source should be specified but not both."
        raise SaltInvocationError(message)
    if source:
        body = __salt__["cp.get_file_str"](
            source, saltenv=__opts__.get("saltenv", "base")
        )
    try:
        return es.index(index=index, doc_type=doc_type, body=body, id=id)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot create document in index {}, server returned code {} with"
            " message {}".format(index, e.status_code, e.error)
        )


def document_delete(index, doc_type, id, hosts=None, profile=None):
    """
    Delete a document from an index

    index
        Index name where the document resides
    doc_type
        Type of the document
    id
        Document identifier

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.document_delete testindex doctype1 AUx-384m0Bug_8U80wQZ
    """
    es = _get_instance(hosts, profile)

    try:
        return es.delete(index=index, doc_type=doc_type, id=id)
    except elasticsearch.exceptions.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot delete document {} in index {}, server returned code {} with"
            " message {}".format(id, index, e.status_code, e.error)
        )


def document_exists(index, id, doc_type="_all", hosts=None, profile=None):
    """
    Return a boolean indicating whether given document exists

    index
        Index name where the document resides
    id
        Document identifier
    doc_type
        Type of the document, use _all to fetch the first document matching the ID across all types

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.document_exists testindex AUx-384m0Bug_8U80wQZ
    """
    es = _get_instance(hosts, profile)

    try:
        return es.exists(index=index, id=id, doc_type=doc_type)
    except elasticsearch.exceptions.NotFoundError:
        return False
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve document {} from index {}, server returned code {} with"
            " message {}".format(id, index, e.status_code, e.error)
        )


def document_get(index, id, doc_type="_all", hosts=None, profile=None):
    """
    Check for the existence of a document and if it exists, return it

    index
        Index name where the document resides
    id
        Document identifier
    doc_type
        Type of the document, use _all to fetch the first document matching the ID across all types

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.document_get testindex AUx-384m0Bug_8U80wQZ
    """
    es = _get_instance(hosts, profile)

    try:
        return es.get(index=index, id=id, doc_type=doc_type)
    except elasticsearch.exceptions.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve document {} from index {}, server returned code {} with"
            " message {}".format(id, index, e.status_code, e.error)
        )


def index_create(index, body=None, hosts=None, profile=None, source=None):
    """
    Create an index

    index
        Index name
    body
        Index definition, such as settings and mappings as defined in https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-create-index.html
    source
        URL to file specifying index definition. Cannot be used in combination with ``body``.

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_create testindex
        salt myminion elasticsearch.index_create testindex2 '{"settings" : {"index" : {"number_of_shards" : 3, "number_of_replicas" : 2}}}'
    """
    es = _get_instance(hosts, profile)
    if source and body:
        message = "Either body or source should be specified but not both."
        raise SaltInvocationError(message)
    if source:
        body = __salt__["cp.get_file_str"](
            source, saltenv=__opts__.get("saltenv", "base")
        )
    try:
        result = es.indices.create(index=index, body=body)
        return result.get("acknowledged", False) and result.get(
            "shards_acknowledged", True
        )
    except elasticsearch.TransportError as e:
        if "index_already_exists_exception" == e.error:
            return True

        raise CommandExecutionError(
            "Cannot create index {}, server returned code {} with message {}".format(
                index, e.status_code, e.error
            )
        )


def index_delete(index, hosts=None, profile=None):
    """
    Delete an index

    index
        Index name

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_delete testindex
    """
    es = _get_instance(hosts, profile)

    try:
        result = es.indices.delete(index=index)

        return result.get("acknowledged", False)
    except elasticsearch.exceptions.NotFoundError:
        return True
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot delete index {}, server returned code {} with message {}".format(
                index, e.status_code, e.error
            )
        )


def index_exists(index, hosts=None, profile=None):
    """
    Return a boolean indicating whether given index exists

    index
        Index name

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_exists testindex
    """
    es = _get_instance(hosts, profile)

    try:
        return es.indices.exists(index=index)
    except elasticsearch.exceptions.NotFoundError:
        return False
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve index {}, server returned code {} with message {}".format(
                index, e.status_code, e.error
            )
        )


def index_get(index, hosts=None, profile=None):
    """
    Check for the existence of an index and if it exists, return it

    index
        Index name

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_get testindex
    """
    es = _get_instance(hosts, profile)

    try:
        return es.indices.get(index=index)
    except elasticsearch.exceptions.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve index {}, server returned code {} with message {}".format(
                index, e.status_code, e.error
            )
        )


def index_open(
    index,
    allow_no_indices=True,
    expand_wildcards="closed",
    ignore_unavailable=True,
    hosts=None,
    profile=None,
):
    """
    .. versionadded:: 2017.7.0

    Open specified index.

    index
        Index to be opened
    allow_no_indices
        Whether to ignore if a wildcard indices expression resolves into no concrete indices. (This includes _all string or when no indices have been specified)
    expand_wildcards
        Whether to expand wildcard expression to concrete indices that are open, closed or both., default ‘closed’, valid choices are: ‘open’, ‘closed’, ‘none’, ‘all’
    ignore_unavailable
        Whether specified concrete indices should be ignored when unavailable (missing or closed)

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_open testindex
    """
    es = _get_instance(hosts, profile)

    try:
        result = es.indices.open(
            index=index,
            allow_no_indices=allow_no_indices,
            expand_wildcards=expand_wildcards,
            ignore_unavailable=ignore_unavailable,
        )

        return result.get("acknowledged", False)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot open index {}, server returned code {} with message {}".format(
                index, e.status_code, e.error
            )
        )


def index_close(
    index,
    allow_no_indices=True,
    expand_wildcards="open",
    ignore_unavailable=True,
    hosts=None,
    profile=None,
):
    """
    .. versionadded:: 2017.7.0

    Close specified index.

    index
        Index to be closed
    allow_no_indices
        Whether to ignore if a wildcard indices expression resolves into no concrete indices. (This includes _all string or when no indices have been specified)
    expand_wildcards
        Whether to expand wildcard expression to concrete indices that are open, closed or both., default ‘open’, valid choices are: ‘open’, ‘closed’, ‘none’, ‘all’
    ignore_unavailable
        Whether specified concrete indices should be ignored when unavailable (missing or closed)

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_close testindex
    """
    es = _get_instance(hosts, profile)

    try:
        result = es.indices.close(
            index=index,
            allow_no_indices=allow_no_indices,
            expand_wildcards=expand_wildcards,
            ignore_unavailable=ignore_unavailable,
        )

        return result.get("acknowledged", False)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot close index {}, server returned code {} with message {}".format(
                index, e.status_code, e.error
            )
        )


def index_get_settings(hosts=None, profile=None, **kwargs):
    """
    .. versionadded:: 3000

    Check for the existence of an index and if it exists, return its settings
    http://www.elastic.co/guide/en/elasticsearch/reference/current/indices-get-settings.html

    index
        (Optional, string) A comma-separated list of index names; use _all or empty string for all indices. Defaults to '_all'.
    name
        (Optional, string) The name of the settings that should be included
    allow_no_indices
        (Optional, boolean) Whether to ignore if a wildcard indices expression resolves into no concrete indices.
        (This includes _all string or when no indices have been specified)
    expand_wildcards
        (Optional, string) Whether to expand wildcard expression to concrete indices that are open, closed or both.
        Valid choices are: ‘open’, ‘closed’, ‘none’, ‘all’
    flat_settings
        (Optional, boolean) Return settings in flat format
    ignore_unavailable
        (Optional, boolean) Whether specified concrete indices should be ignored when unavailable (missing or closed)
    include_defaults
        (Optional, boolean) Whether to return all default setting for each of the indices.
    local
        (Optional, boolean) Return local information, do not retrieve the state from master node

    The defaults settings for the above parameters depend on the API version being used.

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_get_settings index=testindex
    """

    es = _get_instance(hosts, profile)

    # Filtering Salt internal keys
    filtered_kwargs = kwargs.copy()
    for k in kwargs:
        if k.startswith("__"):
            filtered_kwargs.pop(k)

    try:
        return es.indices.get_settings(**filtered_kwargs)
    except elasticsearch.exceptions.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve index settings {}, server returned code {} with message {}".format(
                kwargs, e.status_code, e.error
            )
        )


def index_put_settings(body=None, hosts=None, profile=None, source=None, **kwargs):
    """
    .. versionadded:: 3000

    Update existing index settings
    https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-update-settings.html

    body
        The index settings to be updated.
    source
        URL to file specifying index definition. Cannot be used in combination with ``body``.
    index
        (Optional, string) A comma-separated list of index names; use _all or empty string to perform the operation on all indices
    allow_no_indices
        (Optional, boolean) Whether to ignore if a wildcard indices expression resolves into no concrete indices.
        (This includes _all string or when no indices have been specified)
    expand_wildcards
        (Optional, string) Whether to expand wildcard expression to concrete indices that are open, closed or both.
        Valid choices are: ‘open’, ‘closed’, ‘none’, ‘all’
    flat_settings
        (Optional, boolean) Return settings in flat format (default: false)
    ignore_unavailable
        (Optional, boolean) Whether specified concrete indices should be ignored when unavailable (missing or closed)
    master_timeout
        (Optional, time units) Explicit operation timeout for connection to master node
    preserve_existing
        (Optional, boolean) Whether to update existing settings. If set to true existing settings on an index remain unchanged, the default is false

    The defaults settings for the above parameters depend on the API version being used.

    .. note::
        Elasticsearch time units can be found here:
        https://www.elastic.co/guide/en/elasticsearch/reference/current/common-options.html#time-units

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_put_settings index=testindex body='{"settings" : {"index" : {"number_of_replicas" : 2}}}'
    """
    es = _get_instance(hosts, profile)
    if source and body:
        message = "Either body or source should be specified but not both."
        raise SaltInvocationError(message)
    if source:
        body = __salt__["cp.get_file_str"](
            source, saltenv=__opts__.get("saltenv", "base")
        )

    # Filtering Salt internal keys
    filtered_kwargs = kwargs.copy()
    for k in kwargs:
        if k.startswith("__"):
            filtered_kwargs.pop(k)

    try:
        result = es.indices.put_settings(body=body, **filtered_kwargs)
        return result.get("acknowledged", False)
    except elasticsearch.exceptions.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot update index settings {}, server returned code {} with message {}".format(
                kwargs, e.status_code, e.error
            )
        )


def mapping_create(index, doc_type, body=None, hosts=None, profile=None, source=None):
    """
    Create a mapping in a given index

    index
        Index for the mapping
    doc_type
        Name of the document type
    body
        Mapping definition as specified in https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-put-mapping.html
    source
        URL to file specifying mapping definition. Cannot be used in combination with ``body``.

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.mapping_create testindex user '{ "user" : { "properties" : { "message" : {"type" : "string", "store" : true } } } }'
    """
    es = _get_instance(hosts, profile)
    if source and body:
        message = "Either body or source should be specified but not both."
        raise SaltInvocationError(message)
    if source:
        body = __salt__["cp.get_file_str"](
            source, saltenv=__opts__.get("saltenv", "base")
        )
    try:
        result = es.indices.put_mapping(index=index, doc_type=doc_type, body=body)

        return result.get("acknowledged", False)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot create mapping {}, server returned code {} with message {}".format(
                index, e.status_code, e.error
            )
        )


def mapping_delete(index, doc_type, hosts=None, profile=None):
    """
    Delete a mapping (type) along with its data. As of Elasticsearch 5.0 this is no longer available.

    index
        Index for the mapping
    doc_type
        Name of the document type

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.mapping_delete testindex user
    """
    es = _get_instance(hosts, profile)
    try:
        result = es.indices.delete_mapping(index=index, doc_type=doc_type)

        return result.get("acknowledged", False)
    except elasticsearch.exceptions.NotFoundError:
        return True
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot delete mapping {}, server returned code {} with message {}".format(
                index, e.status_code, e.error
            )
        )
    except AttributeError:
        raise CommandExecutionError("Method is not applicable for Elasticsearch 5.0+")


def mapping_get(index, doc_type, hosts=None, profile=None):
    """
    Retrieve mapping definition of index or index/type

    index
        Index for the mapping
    doc_type
        Name of the document type

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.mapping_get testindex user
    """
    es = _get_instance(hosts, profile)

    try:
        return es.indices.get_mapping(index=index, doc_type=doc_type)
    except elasticsearch.exceptions.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve mapping {}, server returned code {} with message {}".format(
                index, e.status_code, e.error
            )
        )


def index_template_create(name, body=None, hosts=None, profile=None, source=None):
    """
    Create an index template

    name
        Index template name

    body
        Template definition as specified in http://www.elastic.co/guide/en/elasticsearch/reference/current/indices-templates.html

    source
        URL to file specifying template definition. Cannot be used in combination with ``body``.

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_template_create testindex_templ '{ "template": "logstash-*", "order": 1, "settings": { "number_of_shards": 1 } }'
    """
    es = _get_instance(hosts, profile)
    if source and body:
        message = "Either body or source should be specified but not both."
        raise SaltInvocationError(message)
    if source:
        body = __salt__["cp.get_file_str"](
            source, saltenv=__opts__.get("saltenv", "base")
        )
    try:
        result = es.indices.put_template(name=name, body=body)
        return result.get("acknowledged", False)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot create template {}, server returned code {} with message {}".format(
                name, e.status_code, e.error
            )
        )


def index_template_delete(name, hosts=None, profile=None):
    """
    Delete an index template (type) along with its data

    name
        Index template name

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_template_delete testindex_templ user
    """
    es = _get_instance(hosts, profile)
    try:
        result = es.indices.delete_template(name=name)

        return result.get("acknowledged", False)
    except elasticsearch.exceptions.NotFoundError:
        return True
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot delete template {}, server returned code {} with message {}".format(
                name, e.status_code, e.error
            )
        )


def index_template_exists(name, hosts=None, profile=None):
    """
    Return a boolean indicating whether given index template exists

    name
        Index template name

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_template_exists testindex_templ
    """
    es = _get_instance(hosts, profile)
    try:
        return es.indices.exists_template(name=name)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve template {}, server returned code {} with message {}".format(
                name, e.status_code, e.error
            )
        )


def index_template_get(name, hosts=None, profile=None):
    """
    Retrieve template definition of index or index/type

    name
        Index template name

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.index_template_get testindex_templ
    """
    es = _get_instance(hosts, profile)

    try:
        return es.indices.get_template(name=name)
    except elasticsearch.exceptions.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot retrieve template {}, server returned code {} with message {}".format(
                name, e.status_code, e.error
            )
        )


def pipeline_get(id, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Retrieve Ingest pipeline definition. Available since Elasticsearch 5.0.

    id
        Pipeline id

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.pipeline_get mypipeline
    """
    es = _get_instance(hosts, profile)

    try:
        return es.ingest.get_pipeline(id=id)
    except elasticsearch.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot create pipeline {}, server returned code {} with message {}".format(
                id, e.status_code, e.error
            )
        )
    except AttributeError:
        raise CommandExecutionError("Method is applicable only for Elasticsearch 5.0+")


def pipeline_delete(id, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Delete Ingest pipeline. Available since Elasticsearch 5.0.

    id
        Pipeline id

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.pipeline_delete mypipeline
    """
    es = _get_instance(hosts, profile)

    try:
        ret = es.ingest.delete_pipeline(id=id)
        return ret.get("acknowledged", False)
    except elasticsearch.NotFoundError:
        return True
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot delete pipeline {}, server returned code {} with message {}".format(
                id, e.status_code, e.error
            )
        )
    except AttributeError:
        raise CommandExecutionError("Method is applicable only for Elasticsearch 5.0+")


def pipeline_create(id, body, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Create Ingest pipeline by supplied definition. Available since Elasticsearch 5.0.

    id
        Pipeline id
    body
        Pipeline definition as specified in https://www.elastic.co/guide/en/elasticsearch/reference/master/pipeline.html

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.pipeline_create mypipeline '{"description": "my custom pipeline", "processors": [{"set" : {"field": "collector_timestamp_millis", "value": "{{_ingest.timestamp}}"}}]}'
    """
    es = _get_instance(hosts, profile)
    try:
        out = es.ingest.put_pipeline(id=id, body=body)
        return out.get("acknowledged", False)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot create pipeline {}, server returned code {} with message {}".format(
                id, e.status_code, e.error
            )
        )
    except AttributeError:
        raise CommandExecutionError("Method is applicable only for Elasticsearch 5.0+")


def pipeline_simulate(id, body, verbose=False, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Simulate existing Ingest pipeline on provided data. Available since Elasticsearch 5.0.

    id
        Pipeline id
    body
        Pipeline definition as specified in https://www.elastic.co/guide/en/elasticsearch/reference/master/pipeline.html
    verbose
        Specify if the output should be more verbose

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.pipeline_simulate mypipeline '{"docs":[{"_index":"index","_type":"type","_id":"id","_source":{"foo":"bar"}},{"_index":"index","_type":"type","_id":"id","_source":{"foo":"rab"}}]}' verbose=True
    """
    es = _get_instance(hosts, profile)
    try:
        return es.ingest.simulate(id=id, body=body, verbose=verbose)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot simulate pipeline {}, server returned code {} with message {}".format(
                id, e.status_code, e.error
            )
        )
    except AttributeError:
        raise CommandExecutionError("Method is applicable only for Elasticsearch 5.0+")


def search_template_get(id, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Obtain existing search template definition.

    id
        Template ID

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.search_template_get mytemplate
    """
    es = _get_instance(hosts, profile)

    try:
        return es.get_template(id=id)
    except elasticsearch.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot obtain search template {}, server returned code {} with message {}".format(
                id, e.status_code, e.error
            )
        )


def search_template_create(id, body, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Create search template by supplied definition

    id
        Template ID
    body
        Search template definition

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.search_template_create mytemplate '{"template":{"query":{"match":{"title":"{{query_string}}"}}}}'
    """
    es = _get_instance(hosts, profile)

    try:
        result = es.put_template(id=id, body=body)

        return result.get("acknowledged", False)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot create search template {}, server returned code {} with message {}".format(
                id, e.status_code, e.error
            )
        )


def search_template_delete(id, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Delete existing search template definition.

    id
        Template ID

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.search_template_delete mytemplate
    """
    es = _get_instance(hosts, profile)

    try:
        result = es.delete_template(id=id)

        return result.get("acknowledged", False)
    except elasticsearch.NotFoundError:
        return True
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot delete search template {}, server returned code {} with message {}".format(
                id, e.status_code, e.error
            )
        )


def repository_get(name, local=False, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Get existing repository details.

    name
        Repository name
    local
        Retrieve only local information, default is false

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.repository_get testrepo
    """
    es = _get_instance(hosts, profile)

    try:
        return es.snapshot.get_repository(repository=name, local=local)
    except elasticsearch.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot obtain repository {}, server returned code {} with message {}".format(
                name, e.status_code, e.error
            )
        )


def repository_create(name, body, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Create repository for storing snapshots. Note that shared repository paths have to be specified in path.repo Elasticsearch configuration option.

    name
        Repository name
    body
        Repository definition as in https://www.elastic.co/guide/en/elasticsearch/reference/current/modules-snapshots.html

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.repository_create testrepo '{"type":"fs","settings":{"location":"/tmp/test","compress":true}}'
    """
    es = _get_instance(hosts, profile)

    try:
        result = es.snapshot.create_repository(repository=name, body=body)

        return result.get("acknowledged", False)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot create repository {}, server returned code {} with message {}".format(
                name, e.status_code, e.error
            )
        )


def repository_delete(name, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Delete existing repository.

    name
        Repository name

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.repository_delete testrepo
    """
    es = _get_instance(hosts, profile)

    try:
        result = es.snapshot.delete_repository(repository=name)

        return result.get("acknowledged", False)
    except elasticsearch.NotFoundError:
        return True
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot delete repository {}, server returned code {} with message {}".format(
                name, e.status_code, e.error
            )
        )


def repository_verify(name, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Obtain list of cluster nodes which successfully verified this repository.

    name
        Repository name

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.repository_verify testrepo
    """
    es = _get_instance(hosts, profile)

    try:
        return es.snapshot.verify_repository(repository=name)
    except elasticsearch.NotFoundError:
        return None
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot verify repository {}, server returned code {} with message {}".format(
                name, e.status_code, e.error
            )
        )


def snapshot_status(
    repository=None, snapshot=None, ignore_unavailable=False, hosts=None, profile=None
):
    """
    .. versionadded:: 2017.7.0

    Obtain status of all currently running snapshots.

    repository
        Particular repository to look for snapshots
    snapshot
        Snapshot name
    ignore_unavailable
        Ignore unavailable snapshots

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.snapshot_status ignore_unavailable=True
    """
    es = _get_instance(hosts, profile)

    try:
        return es.snapshot.status(
            repository=repository,
            snapshot=snapshot,
            ignore_unavailable=ignore_unavailable,
        )
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot obtain snapshot status, server returned code {} with message {}".format(
                e.status_code, e.error
            )
        )


def snapshot_get(
    repository, snapshot, ignore_unavailable=False, hosts=None, profile=None
):
    """
    .. versionadded:: 2017.7.0

    Obtain snapshot residing in specified repository.

    repository
        Repository name
    snapshot
        Snapshot name, use _all to obtain all snapshots in specified repository
    ignore_unavailable
        Ignore unavailable snapshots

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.snapshot_get testrepo testsnapshot
    """
    es = _get_instance(hosts, profile)

    try:
        return es.snapshot.get(
            repository=repository,
            snapshot=snapshot,
            ignore_unavailable=ignore_unavailable,
        )
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot obtain details of snapshot {} in repository {}, server returned"
            " code {} with message {}".format(
                snapshot, repository, e.status_code, e.error
            )
        )


def snapshot_create(repository, snapshot, body=None, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Create snapshot in specified repository by supplied definition.

    repository
        Repository name
    snapshot
        Snapshot name
    body
        Snapshot definition as in https://www.elastic.co/guide/en/elasticsearch/reference/current/modules-snapshots.html

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.snapshot_create testrepo testsnapshot '{"indices":"index_1,index_2","ignore_unavailable":true,"include_global_state":false}'
    """
    es = _get_instance(hosts, profile)

    try:
        response = es.snapshot.create(
            repository=repository, snapshot=snapshot, body=body
        )

        return response.get("accepted", False)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot create snapshot {} in repository {}, server returned code {} with"
            " message {}".format(snapshot, repository, e.status_code, e.error)
        )


def snapshot_restore(repository, snapshot, body=None, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Restore existing snapshot in specified repository by supplied definition.

    repository
        Repository name
    snapshot
        Snapshot name
    body
        Restore definition as in https://www.elastic.co/guide/en/elasticsearch/reference/current/modules-snapshots.html

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.snapshot_restore testrepo testsnapshot '{"indices":"index_1,index_2","ignore_unavailable":true,"include_global_state":true}'
    """
    es = _get_instance(hosts, profile)

    try:
        response = es.snapshot.restore(
            repository=repository, snapshot=snapshot, body=body
        )

        return response.get("accepted", False)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot restore snapshot {} in repository {}, server returned code {} with"
            " message {}".format(snapshot, repository, e.status_code, e.error)
        )


def snapshot_delete(repository, snapshot, hosts=None, profile=None):
    """
    .. versionadded:: 2017.7.0

    Delete snapshot from specified repository.

    repository
        Repository name
    snapshot
        Snapshot name

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.snapshot_delete testrepo testsnapshot
    """
    es = _get_instance(hosts, profile)

    try:
        result = es.snapshot.delete(repository=repository, snapshot=snapshot)

        return result.get("acknowledged", False)
    except elasticsearch.NotFoundError:
        return True
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot delete snapshot {} from repository {}, server returned code {} with"
            " message {}".format(snapshot, repository, e.status_code, e.error)
        )


def flush_synced(hosts=None, profile=None, **kwargs):
    """
    .. versionadded:: 3000

    Perform a normal flush, then add a generated unique marker (sync_id) to all shards.
    http://www.elastic.co/guide/en/elasticsearch/reference/current/indices-synced-flush.html

    index
        (Optional, string) A comma-separated list of index names; use _all or empty string for all indices. Defaults to '_all'.

    ignore_unavailable
        (Optional, boolean) If true, missing or closed indices are not included in the response. Defaults to false.

    allow_no_indices
        (Optional, boolean) If true, the request does not return an error if a wildcard expression or _all value retrieves only missing or closed indices.
        This parameter also applies to index aliases that point to a missing or closed index.

    expand_wildcards
        (Optional, string) Controls what kind of indices that wildcard expressions can expand to.

        Valid values are::

            all - Expand to open and closed indices.
            open - Expand only to open indices.
            closed - Expand only to closed indices.
            none - Wildcard expressions are not accepted.

    The defaults settings for the above parameters depend on the API being used.

    CLI Example:

    .. code-block:: bash

        salt myminion elasticsearch.flush_synced index='index1,index2' ignore_unavailable=True allow_no_indices=True expand_wildcards='all'
    """
    es = _get_instance(hosts, profile)

    try:
        return es.indices.flush_synced(kwargs)
    except elasticsearch.TransportError as e:
        raise CommandExecutionError(
            "Cannot flush synced, server returned code {} with message {}".format(
                e.status_code, e.error
            )
        )



def test_syndic(salt_cli, minion):
    ret = salt_cli.run("test.ping", minion_tgt="*", _timeout=15)
    assert ret.data == {
        "syndic": True,
        "minion": True,
    }


def test_syndic_target_single_minion(salt_cli, minion):
    """
    Test that the salt CLI exits after getting all returns + syndic_wait time
    rather than waiting for the full gather_job_timeout set high in conftest.
    """
    ret = salt_cli.run("test.ping", minion_tgt="minion", _timeout=15)
    assert ret.data is True



"""
Installation of PHP Extensions Using pecl
=========================================

These states manage the installed pecl extensions. Note that php-pear must be
installed for these states to be available, so pecl states should include a
requisite to a pkg.installed state for the package which provides pecl
(``php-pear`` in most cases). Example:

.. code-block:: yaml

    php-pear:
      pkg.installed

    mongo:
      pecl.installed:
        - require:
          - pkg: php-pear
"""


def __virtual__():
    """
    Only load if the pecl module is available in __salt__
    """
    if "pecl.list" in __salt__:
        return "pecl"
    return (False, "pecl module could not be loaded")


def installed(
    name, version=None, defaults=False, force=False, preferred_state="stable"
):
    """
    .. versionadded:: 0.17.0

    Make sure that a pecl extension is installed.

    name
        The pecl extension name to install

    version
        The pecl extension version to install. This option may be
        ignored to install the latest stable version.

    defaults
        Use default answers for extensions such as pecl_http which ask
        questions before installation. Without this option, the pecl.installed
        state will hang indefinitely when trying to install these extensions.

    force
        Whether to force the installed version or not

    preferred_state
        The pecl extension state to install
    """
    # Check to see if we have a designated version
    if not isinstance(version, str) and version is not None:
        version = str(version)

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    if "/" in name:
        channel, package = name.split("/")
    else:
        channel = None
        package = name
    installed_pecls = __salt__["pecl.list"](channel)

    if package in installed_pecls:
        # The package is only installed if version is absent or matches
        if (
            version is None or version in installed_pecls[package]
        ) and preferred_state in installed_pecls[package]:
            ret["result"] = True
            ret["comment"] = f"Pecl extension {name} is already installed."
            return ret

    if version is not None:
        # Modify the name to include the version and proceed.
        name = f"{name}-{version}"

    if __opts__["test"]:
        ret["comment"] = f"Pecl extension {name} would have been installed"
        return ret
    if __salt__["pecl.install"](
        name, defaults=defaults, force=force, preferred_state=preferred_state
    ):
        ret["result"] = True
        ret["changes"][name] = "Installed"
        ret["comment"] = f"Pecl extension {name} was successfully installed"
    else:
        ret["result"] = False
        ret["comment"] = f"Could not install pecl extension {name}."

    return ret


def removed(name):
    """
    Make sure that a pecl extension is not installed.

    name
        The pecl extension name to uninstall
    """
    ret = {"name": name, "result": None, "comment": "", "changes": {}}
    if name not in __salt__["pecl.list"]():
        ret["result"] = True
        ret["comment"] = f"Pecl extension {name} is not installed."
        return ret

    if __opts__["test"]:
        ret["comment"] = f"Pecl extension {name} would have been removed"
        return ret
    if __salt__["pecl.uninstall"](name):
        ret["result"] = True
        ret["changes"][name] = "Removed"
        ret["comment"] = f"Pecl extension {name} was successfully removed."
    else:
        ret["result"] = False
        ret["comment"] = f"Could not remove pecl extension {name}."
    return ret



# pylint: disable=unexpected-keyword-arg
import pytest

import salt.modules.purefa as purefa
from tests.support.mock import MagicMock, call, create_autospec, patch


@pytest.fixture
def configure_loader_modules():
    fake_pure_storage = MagicMock()
    fake_pure_storage.PureError = Exception
    return {
        purefa: {
            "_get_host": create_autospec(purefa._get_host, return_value=None),
            "purestorage": fake_pure_storage,
        }
    }


@pytest.fixture
def patch_get_system():
    with patch("salt.modules.purefa._get_system", autospec=True) as fake_sys:
        yield fake_sys


@pytest.fixture
def fake_set_host(patch_get_system):
    fake_set_host = MagicMock()
    patch_get_system.return_value.set_host = fake_set_host
    yield fake_set_host


@pytest.fixture
def fake_delete_host(patch_get_system):
    fake_delete_host = MagicMock()
    patch_get_system.return_value.delete_host = fake_delete_host
    yield fake_delete_host


@pytest.mark.parametrize("nqn", [None, "", [], ()])
def test_when_nqn_is_not_anything_set_host_should_not_have_addnqnlist_called(
    nqn, fake_set_host
):
    purefa.host_create("fnord", nqn=nqn)

    for call in fake_set_host.mock_calls:
        assert "addnqnlist" not in call.kwargs


def test_when_nqn_is_provided_and_adding_is_successful_then_set_host_should_have_addqnlist(
    fake_set_host,
):
    nqn = "fnord"
    host = "fnord-host"
    expected_calls = [call(host, addnqnlist=[nqn])]
    purefa.host_create(host, nqn=nqn)

    fake_set_host.assert_has_calls(expected_calls)


def test_when_nqn_is_provided_and_adding_is_successful_then_result_should_be_True(
    fake_set_host,
):

    result = purefa.host_create("fnord", nqn="fnordqn")

    assert result is True


def test_when_nqn_is_provided_but_set_host_fails_then_creation_should_be_rolled_back(
    fake_set_host, fake_delete_host
):
    expected_host = "fnord"
    fake_set_host.side_effect = purefa.purestorage.PureError("oops!")
    result = purefa.host_create(expected_host, nqn="badness or whatever")

    assert result is False
    fake_delete_host.assert_called_with(expected_host)


@pytest.mark.parametrize("nqn", [None, "", [], ()])
def test_when_nqn_is_not_then_host_update_should_not_call_set_host(fake_set_host, nqn):
    # if _get_host is None then there's no host to update, is there?
    purefa._get_host.return_value = True

    purefa.host_update("fnord", nqn=nqn)

    for call in fake_set_host.mock_calls:
        assert "addnqnlist" not in call.kwargs


def test_when_nqn_is_correctly_provided_it_should_be_set_on_the_host(fake_set_host):
    purefa._get_host.return_value = True
    expected_calls = [call("fnord", addnqnlist=["roscivs"])]

    purefa.host_update("fnord", nqn="roscivs")

    fake_set_host.assert_has_calls(expected_calls)


def test_when_nqn_is_correctly_provided_result_should_be_True(fake_set_host):
    purefa._get_host.return_value = True

    result = purefa.host_update("fnord", nqn="roscivs")

    assert result is True


def test_when_nqn_fails_result_should_be_False(fake_set_host):
    purefa._get_host.return_value = True
    fake_set_host.side_effect = purefa.purestorage.PureError("oops!")

    result = purefa.host_update("fnord", nqn="roscivs")

    assert result is False



"""
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>

    Test cases for salt.modules.osquery
"""

import pytest

import salt.modules.osquery as osquery
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {osquery: {}}


def test_version():
    """
    Test the version returned from OSQuery
    """
    _table_attrs_results = [
        "pid",
        "uuid",
        "instance_id",
        "version",
        "config_hash",
        "config_valid",
        "extensions",
        "build_platform",
        "build_distro",
        "start_time",
        "watcher",
    ]

    _os_query_results = {"data": [{"version": "2.6.1"}], "result": True}

    with patch.object(
        osquery, "_table_attrs", MagicMock(return_value=_table_attrs_results)
    ):
        with patch.object(
            osquery, "_osquery", MagicMock(return_value=_os_query_results)
        ):
            assert osquery.version() == "2.6.1"


def test_deb_packages():
    """
    Test the results returned from the deb_packages function
    """
    _os_query_results = {
        "data": [
            {
                "arch": "amd64",
                "name": "accountsservice",
                "revision": "1",
                "size": "451",
                "source": "",
                "version": "0.6.45-1",
            },
            {
                "arch": "amd64",
                "name": "acetoneiso",
                "revision": "2+b2",
                "size": "1820",
                "source": "acetoneiso (2.4-2)",
                "version": "2.4-2+b2",
            },
            {
                "arch": "amd64",
                "name": "acl",
                "revision": "3+b1",
                "size": "200",
                "source": "acl (2.2.52-3)",
                "version": "2.2.52-3+b1",
            },
            {
                "arch": "amd64",
                "name": "adb",
                "revision": "2",
                "size": "189",
                "source": "android-platform-system-core",
                "version": "1: 7.0.0+r33-2",
            },
        ],
        "result": True,
    }
    with patch.object(osquery, "_osquery", MagicMock(return_value=_os_query_results)):
        with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
            assert osquery.deb_packages() == _os_query_results


def test_deb_packages_with_attrs():
    """
    Test the results returned from the deb_packages function
    with attributes
    """
    _table_attrs_results = ["name", "version", "source", "size", "arch", "revision"]

    _os_query_results = {
        "data": [
            {"name": "accountsservice", "version": "0.6.45-1"},
            {"name": "acetoneiso", "version": "2.4-2+b2"},
            {"name": "acl", "version": "2.2.52-3+b1"},
            {"name": "adb", "version": "1: 7.0.0+r33-2"},
        ],
        "result": True,
    }

    with patch.object(
        osquery, "_table_attrs", MagicMock(return_value=_table_attrs_results)
    ):
        with patch.object(
            osquery, "_osquery", MagicMock(return_value=_os_query_results)
        ):
            with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
                assert (
                    osquery.deb_packages(attrs=["name", "version"]) == _os_query_results
                )


def test_kernel_modules():
    """
    Test the results returned from the kernel_modules function
    """
    _os_query_results = {
        "data": [
            {
                "address": "0xffffffffc14f2000",
                "name": "nls_utf8",
                "size": "16384",
                "status": "Live",
                "used_by": "-",
            },
            {
                "address": "0xffffffffc1599000",
                "name": "udf",
                "size": "90112",
                "status": "Live",
                "used_by": "-",
            },
            {
                "address": "0xffffffffc14b5000",
                "name": "crc_itu_t",
                "size": "16384",
                "status": "Live",
                "used_by": "udf",
            },
        ],
        "result": True,
    }
    with patch.object(osquery, "_osquery", MagicMock(return_value=_os_query_results)):
        with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
            assert osquery.kernel_modules() == _os_query_results


def test_kernel_modules_with_attrs():
    """
    Test the results returned from the kernel_modules function
    with attributes
    """
    _table_attrs_results = ["address", "name", "size", "status", "used_by"]

    _os_query_results = {
        "data": [
            {"name": "nls_utf8", "status": "Live"},
            {"name": "udf", "status": "Live"},
            {"name": "crc_itu_t", "status": "Live"},
        ],
        "result": True,
    }
    with patch.object(
        osquery, "_table_attrs", MagicMock(return_value=_table_attrs_results)
    ):
        with patch.object(
            osquery, "_osquery", MagicMock(return_value=_os_query_results)
        ):
            with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
                assert (
                    osquery.kernel_modules(attrs=["name", "status"])
                    == _os_query_results
                )


def test_osquery_info():
    """
    Test the results returned from the kernel_modules function
    with attributes
    """
    _table_attrs_results = [
        "pid",
        "uuid",
        "instance_id",
        "version",
        "config_hash",
        "config_valid",
        "extensions",
        "build_platform",
        "build_distro",
        "start_time",
        "watcher",
    ]

    _os_query_results = {
        "data": [
            {
                "build_platform": "ubuntu",
                "start_time": "1514484833",
                "uuid": "D31FD400-7277-11E3-ABA6-B8AEED7E173B",
                "build_distro": "xenial",
                "pid": "24288",
                "watcher": "-1",
                "instance_id": "dff196b0-5c91-4105-962b-28660d7aa282",
                "version": "2.6.1",
                "extensions": "inactive",
                "config_valid": "0",
                "config_hash": "",
            }
        ],
        "result": True,
    }

    with patch.object(
        osquery, "_table_attrs", MagicMock(return_value=_table_attrs_results)
    ):
        with patch.object(
            osquery, "_osquery", MagicMock(return_value=_os_query_results)
        ):
            with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
                assert osquery.osquery_info() == _os_query_results


def test_osquery_info_with_attrs():
    """
    Test the results returned from the kernel_modules function
    with attributes
    """
    _table_attrs_results = [
        "pid",
        "uuid",
        "instance_id",
        "version",
        "config_hash",
        "config_valid",
        "extensions",
        "build_platform",
        "build_distro",
        "start_time",
        "watcher",
    ]

    _os_query_results = {
        "data": [{"build_platform": "ubuntu", "start_time": "1514484833"}],
        "result": True,
    }

    with patch.object(
        osquery, "_table_attrs", MagicMock(return_value=_table_attrs_results)
    ):
        with patch.object(
            osquery, "_osquery", MagicMock(return_value=_os_query_results)
        ):
            with patch.dict(osquery.__grains__, {"os_family": "Debian"}):
                assert (
                    osquery.osquery_info(attrs=["build_platform", "start_time"])
                    == _os_query_results
                )



import logging
import sys
import time

import packaging.version
import psutil
import pytest
from pytestskipmarkers.utils import platform

log = logging.getLogger(__name__)


@pytest.fixture
def salt_systemd_setup(
    salt_call_cli,
    install_salt,
):
    """
    Fixture to set systemd for salt packages to enabled and active
    Note: assumes Salt packages already installed
    """
    # ensure known state, enabled and active
    test_list = ["salt-api", "salt-minion", "salt-master"]
    for test_item in test_list:
        test_cmd = f"systemctl enable {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
        assert ret.returncode == 0

        test_cmd = f"systemctl restart {test_item}"
        ret = salt_call_cli.run("--local", "cmd.run", test_cmd)
        assert ret.returncode == 0


def salt_test_upgrade(
    salt_call_cli,
    install_salt,
):
    """
    Test upgrade of Salt packages for Minion and Master
    """
    log.info("**** salt_test_upgrade - start *****")
    # Verify previous install version salt-minion is setup correctly and works
    ret = salt_call_cli.run("--local", "test.version")
    assert ret.returncode == 0
    installed_minion_version = packaging.version.parse(ret.data)
    assert installed_minion_version < packaging.version.parse(
        install_salt.artifact_version
    )

    # Verify previous install version salt-master is setup correctly and works
    bin_file = "salt"
    if sys.platform == "win32":
        bin_file = "salt-call.exe"
    ret = install_salt.proc.run(bin_file, "--version")
    assert ret.returncode == 0
    assert packaging.version.parse(
        ret.stdout.strip().split()[1]
    ) < packaging.version.parse(install_salt.artifact_version)

    # Verify there is a running minion and master by getting there PIDs
    if platform.is_windows():
        process_master_name = "cli_salt_master.py"
        process_minion_name = "salt-minion.exe"
    else:
        process_master_name = "salt-master"
        process_minion_name = "salt-minion"

    old_minion_pids = _get_running_named_salt_pid(process_minion_name)
    old_master_pids = _get_running_named_salt_pid(process_master_name)
    assert old_minion_pids
    assert old_master_pids

    # Upgrade Salt (inc. minion, master, etc.) from previous version and test
    install_salt.install(upgrade=True)

    start = time.monotonic()
    while True:
        ret = salt_call_cli.run("--local", "test.version", _timeout=10)
        if ret.returncode == 0:
            break
        if time.monotonic() - start > 60:
            break

    ret = salt_call_cli.run("--local", "test.version")
    assert ret.returncode == 0

    installed_minion_version = packaging.version.parse(ret.data)
    assert installed_minion_version == packaging.version.parse(
        install_salt.artifact_version
    )

    ret = install_salt.proc.run(bin_file, "--version")
    assert ret.returncode == 0
    assert packaging.version.parse(
        ret.stdout.strip().split()[1]
    ) == packaging.version.parse(install_salt.artifact_version)

    # Verify there is a new running minion and master by getting their PID and comparing them
    # with previous PIDs from before the upgrade

    new_minion_pids = _get_running_named_salt_pid(process_minion_name)
    new_master_pids = _get_running_named_salt_pid(process_master_name)

    if sys.platform == "linux" and install_salt.distro_id not in ("ubuntu", "debian"):
        assert new_minion_pids
        assert new_master_pids
        assert new_minion_pids != old_minion_pids
        assert new_master_pids != old_master_pids

    log.info("**** salt_test_upgrade - end *****")


def _get_running_named_salt_pid(process_name):

    # need to check all of command line for salt-minion, salt-master, for example: salt-minion
    #
    # Linux: psutil process name only returning first part of the command '/opt/saltstack/'
    # Linux: ['/opt/saltstack/salt/bin/python3.10 /usr/bin/salt-minion MultiMinionProcessManager MinionProcessManager']
    #
    # MacOS: psutil process name only returning last part of the command '/opt/salt/bin/python3.10', that is 'python3.10'
    # MacOS: ['/opt/salt/bin/python3.10 /opt/salt/salt-minion', '']

    pids = []
    for proc in psutil.process_iter():
        try:
            cmdl_strg = " ".join(str(element) for element in proc.cmdline())
        except psutil.AccessDenied:
            continue
        if process_name in cmdl_strg:
            pids.append(proc.pid)

    return pids


def _get_installed_salt_packages():
    """
    Get list of installed Salt packages on Windows via registry.
    Returns list of tuples: (name, version)
    """
    if not platform.is_windows():
        return []

    import subprocess

    cmd = [
        "powershell",
        "-Command",
        (
            "Get-ItemProperty HKLM:\\Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall\\* | "
            "Where-Object { $_.DisplayName -like '*Salt*' } | "
            "Select-Object DisplayName, DisplayVersion | "
            'ForEach-Object { "$($_.DisplayName)|$($_.DisplayVersion)" }'
        ),
    ]

    result = subprocess.run(cmd, capture_output=True, text=True, check=False)
    if result.returncode != 0:
        log.warning("Failed to query installed packages: %s", result.stderr)
        return []

    packages = []
    for line in result.stdout.strip().split("\n"):
        line = line.strip()
        if line and "|" in line:
            name, version = line.split("|", 1)
            packages.append((name.strip(), version.strip()))

    return packages


def test_salt_upgrade(salt_call_cli, install_salt, debian_disable_policy_rcd):
    """
    Test an upgrade of Salt, Minion and Master
    """
    if not install_salt.upgrade:
        pytest.skip("Not testing an upgrade, do not run")

    original_py_version = install_salt.package_python_version()

    # Test pip install before an upgrade
    dep = "PyGithub==1.56.0"
    install = salt_call_cli.run("--local", "pip.install", dep)
    assert install.returncode == 0

    # Verify we can use the module dependent on the installed package
    repo = "https://github.com/saltstack/salt.git"
    use_lib = salt_call_cli.run("--local", "github.get_repo_info", repo)
    assert "Authentication information could" in use_lib.stderr

    # perform Salt package upgrade test
    salt_test_upgrade(salt_call_cli, install_salt)

    # Verify only one Salt package is installed after upgrade (Windows)
    if platform.is_windows():
        installed_packages = _get_installed_salt_packages()
        log.info("Installed Salt packages after upgrade: %s", installed_packages)
        assert len(installed_packages) == 1, (
            f"Expected 1 Salt package after upgrade, found {len(installed_packages)}: "
            f"{installed_packages}"
        )
        package_name, package_version = installed_packages[0]
        log.info(
            "Verified single package: %s version %s", package_name, package_version
        )

    new_py_version = install_salt.package_python_version()
    if new_py_version == original_py_version:
        # test pip install after an upgrade
        use_lib = salt_call_cli.run("--local", "github.get_repo_info", repo)
        assert "Authentication information could" in use_lib.stderr



"""
Test the napalm_formula execution module.
"""

import textwrap

import pytest

import salt.modules.napalm_formula as napalm_formula
from salt.utils.immutabletypes import freeze
from tests.support.mock import MagicMock, patch


@pytest.fixture
def set_model():
    return freeze(
        {
            "interfaces": {
                "interface": {
                    "Ethernet1": {
                        "config": {
                            "name": "Ethernet1",
                            "description": "Interface Ethernet1",
                        },
                        "subinterfaces": {
                            "subinterface": {
                                "0": {
                                    "config": {
                                        "index": 0,
                                        "description": "Subinterface Ethernet1.0",
                                    }
                                },
                                "100": {
                                    "config": {
                                        "index": 100,
                                        "description": "Subinterface Ethernet1.100",
                                    }
                                },
                                "900": {
                                    "config": {
                                        "index": 900,
                                        "description": "Subinterface Ethernet1.900",
                                    }
                                },
                            }
                        },
                    },
                    "Ethernet2": {
                        "config": {
                            "name": "Ethernet2",
                            "description": "Interface Ethernet2",
                        },
                        "subinterfaces": {
                            "subinterface": {
                                "400": {
                                    "config": {
                                        "index": 400,
                                        "description": "Subinterface Ethernet2.400",
                                    }
                                }
                            }
                        },
                    },
                }
            }
        }
    )


@pytest.fixture
def set_defaults():
    return freeze(
        {
            "interfaces": {
                "interface": {
                    "*": {
                        "config": {"mtu": 2048, "enabled": True},
                        "subinterfaces": {
                            "subinterface": {"*": {"config": {"enabled": True}}}
                        },
                    }
                }
            }
        }
    )


@pytest.fixture
def configure_loader_modules():
    return {napalm_formula: {"__grains__": {"os": "eos"}}}


def test_container_path(set_model):
    paths = [
        "interfaces:interface:Ethernet1:config",
        "interfaces:interface:Ethernet1:subinterfaces:subinterface:0:config",
        "interfaces:interface:Ethernet1:subinterfaces:subinterface:100:config",
        "interfaces:interface:Ethernet2:subinterfaces:subinterface:400:config",
        "interfaces:interface:Ethernet1:subinterfaces:subinterface:900:config",
        "interfaces:interface:Ethernet2:config",
    ]
    with patch("salt.utils.napalm.is_proxy", MagicMock(return_value=True)):
        ret = napalm_formula.container_path(set_model.copy())
        assert set(ret) == set(paths)


def test_setval():
    with patch("salt.utils.napalm.is_proxy", MagicMock(return_value=True)):
        dict_ = {"foo": {"bar": {"baz": True}}}
        assert dict_ == napalm_formula.setval("foo:bar:baz", True)


def test_defaults(set_model, set_defaults):
    expected_result = {
        "interfaces": {
            "interface": {
                "Ethernet1": {
                    "config": {
                        "name": "Ethernet1",
                        "description": "Interface Ethernet1",
                        "mtu": 2048,
                        "enabled": True,
                    },
                    "subinterfaces": {
                        "subinterface": {
                            "0": {
                                "config": {
                                    "index": 0,
                                    "description": "Subinterface Ethernet1.0",
                                    "enabled": True,
                                }
                            },
                            "100": {
                                "config": {
                                    "index": 100,
                                    "description": "Subinterface Ethernet1.100",
                                    "enabled": True,
                                }
                            },
                            "900": {
                                "config": {
                                    "index": 900,
                                    "description": "Subinterface Ethernet1.900",
                                    "enabled": True,
                                }
                            },
                        }
                    },
                },
                "Ethernet2": {
                    "config": {
                        "name": "Ethernet2",
                        "description": "Interface Ethernet2",
                        "mtu": 2048,
                        "enabled": True,
                    },
                    "subinterfaces": {
                        "subinterface": {
                            "400": {
                                "config": {
                                    "index": 400,
                                    "description": "Subinterface Ethernet2.400",
                                    "enabled": True,
                                }
                            }
                        }
                    },
                },
            }
        }
    }
    with patch("salt.utils.napalm.is_proxy", MagicMock(return_value=True)):
        ret = napalm_formula.defaults(set_model.copy(), set_defaults.copy())
        assert ret == expected_result


def test_render_field():
    config = {"description": "Interface description"}
    ret = napalm_formula.render_field(config, "description", quotes=True)
    assert ret == 'description "Interface description"'


def test_render_field_junos():
    config = {"description": "Interface description"}
    with patch.dict(napalm_formula.__grains__, {"os": "junos"}):
        ret = napalm_formula.render_field(config, "description", quotes=True)
        assert ret == 'description "Interface description";'


def test_render_fields():
    config = {"mtu": 2048, "description": "Interface description"}
    expected_render = textwrap.dedent(
        '''\
            mtu "2048"
            description "Interface description"'''
    )
    ret = napalm_formula.render_fields(config, "mtu", "description", quotes=True)
    assert ret == expected_render



"""
Module for managing SNMP service settings on Windows servers.

"""


def __virtual__():
    """
    Load only on minions that have the win_snmp module.
    """
    if "win_snmp.get_agent_settings" in __salt__:
        return True
    return (False, "win_snmp module could not be loaded")


def agent_settings(name, contact, location, services=None):
    """
    Manage the SNMP sysContact, sysLocation, and sysServices settings.

    :param str contact: The SNMP contact.
    :param str location: The SNMP location.
    :param str services: A list of selected services.

    Example of usage:

    .. code-block:: yaml

        snmp-agent-settings:
            win_snmp.agent_settings:
                - contact: Test Contact
                - location: Test Location
                - services:
                    - Physical
                    - Internet
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": None}

    ret_settings = {"changes": dict(), "failures": dict()}

    if not services:
        services = ["None"]

    # Filter services for unique items, and sort them for comparison purposes.
    services = sorted(set(services))

    settings = {"contact": contact, "location": location, "services": services}

    current_settings = __salt__["win_snmp.get_agent_settings"]()

    for setting in settings:
        if str(settings[setting]) != str(current_settings[setting]):
            ret_settings["changes"][setting] = {
                "old": current_settings[setting],
                "new": settings[setting],
            }
    if not ret_settings["changes"]:
        ret["comment"] = "Agent settings already contain the provided values."
        ret["result"] = True
        return ret
    elif __opts__["test"]:
        ret["comment"] = "Agent settings will be changed."
        ret["changes"] = ret_settings
        return ret

    __salt__["win_snmp.set_agent_settings"](**settings)
    new_settings = __salt__["win_snmp.get_agent_settings"]()

    for setting in settings:
        if settings[setting] != new_settings[setting]:
            ret_settings["failures"][setting] = {
                "old": current_settings[setting],
                "new": new_settings[setting],
            }
            ret_settings["changes"].pop(setting, None)

    if ret_settings["failures"]:
        ret["comment"] = "Some agent settings failed to change."
        ret["changes"] = ret_settings
        ret["result"] = False
    else:
        ret["comment"] = "Set agent settings to contain the provided values."
        ret["changes"] = ret_settings["changes"]
        ret["result"] = True
    return ret


def auth_traps_enabled(name, status=True):
    """
    Manage the sending of authentication traps.

    :param bool status: The enabled status.

    Example of usage:

    .. code-block:: yaml

        snmp-auth-traps:
            win_snmp.auth_traps_enabled:
                - status: True
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": None}

    vname = "EnableAuthenticationTraps"
    current_status = __salt__["win_snmp.get_auth_traps_enabled"]()

    if status == current_status:
        ret["comment"] = f"{vname} already contains the provided value."
        ret["result"] = True
    elif __opts__["test"]:
        ret["comment"] = f"{vname} will be changed."
        ret["changes"] = {"old": current_status, "new": status}
    else:
        ret["comment"] = f"Set {vname} to contain the provided value."
        ret["changes"] = {"old": current_status, "new": status}
        ret["result"] = __salt__["win_snmp.set_auth_traps_enabled"](status=status)

    return ret


def community_names(name, communities=None):
    """
    Manage the SNMP accepted community names and their permissions.

    :param str communities: A dictionary of SNMP communities and permissions.

    Example of usage:

    .. code-block:: yaml

        snmp-community-names:
            win_snmp.community_names:
                - communities:
                    TestCommunity: Read Only
                    OtherCommunity: Read Write
    """
    ret = {"name": name, "changes": dict(), "comment": "", "result": None}

    ret_communities = {"changes": dict(), "failures": dict()}

    if not communities:
        communities = dict()

    current_communities = __salt__["win_snmp.get_community_names"]()

    # Note any existing communities that should be removed.
    for current_vname in current_communities:
        if current_vname not in communities:
            ret_communities["changes"][current_vname] = {
                "old": current_communities[current_vname],
                "new": None,
            }

    # Note any new communities or existing communities that should be changed.
    for vname in communities:
        current_vdata = None
        if vname in current_communities:
            current_vdata = current_communities[vname]
        if communities[vname] != current_vdata:
            ret_communities["changes"][vname] = {
                "old": current_vdata,
                "new": communities[vname],
            }

    if not ret_communities["changes"]:
        ret["comment"] = "Communities already contain the provided values."
        ret["result"] = True
        return ret
    elif __opts__["test"]:
        ret["comment"] = "Communities will be changed."
        ret["changes"] = ret_communities
        return ret

    __salt__["win_snmp.set_community_names"](communities=communities)
    new_communities = __salt__["win_snmp.get_community_names"]()

    # Verify that any communities that needed to be removed were removed.
    for new_vname in new_communities:
        if new_vname not in communities:
            ret_communities["failures"][new_vname] = {
                "old": current_communities[new_vname],
                "new": new_communities[new_vname],
            }
            ret_communities["changes"].pop(new_vname, None)

    # Verify that any new communities or existing communities that
    # needed to be changed were changed.
    for vname in communities:
        new_vdata = None
        if vname in new_communities:
            new_vdata = new_communities[vname]
        if communities[vname] != new_vdata:
            ret_communities["failures"][vname] = {
                "old": current_communities[vname],
                "new": new_vdata,
            }
            ret_communities["changes"].pop(vname, None)

    if ret_communities["failures"]:
        ret["comment"] = "Some communities failed to change."
        ret["changes"] = ret_communities
        ret["result"] = False
    else:
        ret["comment"] = "Set communities to contain the provided values."
        ret["changes"] = ret_communities["changes"]
        ret["result"] = True
    return ret



"""
Utility functions for zfs

These functions are for dealing with type conversion and basic execution

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       salt.utils.stringutils, salt.ext, salt.module.cmdmod
:platform:      illumos,freebsd,linux

.. versionadded:: 2018.3.1

"""

import logging
import math
import os
import re
from collections import OrderedDict
from numbers import Number

import salt.modules.cmdmod
from salt.utils.decorators import memoize as real_memoize
from salt.utils.stringutils import to_num as str_to_num

# Size conversion data
re_zfs_size = re.compile(r"^(\d+|\d+(?=\d*)\.\d+)([KkMmGgTtPpEe][Bb]?)$")
zfs_size = ["K", "M", "G", "T", "P", "E"]

log = logging.getLogger(__name__)


def _check_retcode(cmd):
    """
    Simple internal wrapper for cmdmod.retcode
    """
    return (
        salt.modules.cmdmod.retcode(cmd, output_loglevel="quiet", ignore_retcode=True)
        == 0
    )


def _exec(**kwargs):
    """
    Simple internal wrapper for cmdmod.run
    """
    if "ignore_retcode" not in kwargs:
        kwargs["ignore_retcode"] = True
    if "output_loglevel" not in kwargs:
        kwargs["output_loglevel"] = "quiet"
    return salt.modules.cmdmod.run_all(**kwargs)


def _merge_last(values, merge_after, merge_with=" "):
    """
    Merge values all values after X into the last value
    """
    if len(values) > merge_after:
        values = values[0 : (merge_after - 1)] + [
            merge_with.join(values[(merge_after - 1) :])
        ]

    return values


def _property_normalize_name(name):
    """
    Normalizes property names
    """
    if "@" in name:
        name = name[: name.index("@") + 1]
    return name


def _property_detect_type(name, values):
    """
    Detect the datatype of a property
    """
    value_type = "str"
    if values.startswith("on | off"):
        value_type = "bool"
    elif values.startswith("yes | no"):
        value_type = "bool_alt"
    elif values in ["<size>", "<size> | none"]:
        value_type = "size"
    elif values in ["<count>", "<count> | none", "<guid>"]:
        value_type = "numeric"
    elif name in ["sharenfs", "sharesmb", "canmount"]:
        value_type = "bool"
    elif name in ["version", "copies"]:
        value_type = "numeric"
    return value_type


def _property_create_dict(header, data):
    """
    Create a property dict
    """
    prop = dict(zip(header, _merge_last(data, len(header))))
    prop["name"] = _property_normalize_name(prop["property"])
    prop["type"] = _property_detect_type(prop["name"], prop["values"])
    prop["edit"] = from_bool(prop["edit"])
    if "inherit" in prop:
        prop["inherit"] = from_bool(prop["inherit"])
    del prop["property"]
    return prop


def _property_parse_cmd(cmd, alias=None):
    """
    Parse output of zpool/zfs get command
    """
    if not alias:
        alias = {}
    properties = {}

    # NOTE: append get to command
    if cmd[-3:] != "get":
        cmd += " get"

    # NOTE: parse output
    prop_hdr = []
    for prop_data in _exec(cmd=cmd)["stderr"].split("\n"):
        # NOTE: make the line data more manageable
        prop_data = prop_data.lower().split()

        # NOTE: skip empty lines
        if not prop_data:
            continue
        # NOTE: parse header
        elif prop_data[0] == "property":
            prop_hdr = prop_data
            continue
        # NOTE: skip lines after data
        elif not prop_hdr or prop_data[1] not in ["no", "yes"]:
            continue

        # NOTE: create property dict
        prop = _property_create_dict(prop_hdr, prop_data)

        # NOTE: add property to dict
        properties[prop["name"]] = prop
        if prop["name"] in alias:
            properties[alias[prop["name"]]] = prop

        # NOTE: cleanup some duplicate data
        del prop["name"]
    return properties


def _auto(direction, name, value, source="auto", convert_to_human=True):
    """
    Internal magic for from_auto and to_auto
    """
    # NOTE: check direction
    if direction not in ["to", "from"]:
        return value

    # NOTE: collect property data
    props = property_data_zpool()
    if source == "zfs":
        props = property_data_zfs()
    elif source == "auto":
        props.update(property_data_zfs())

    # NOTE: figure out the conversion type
    value_type = props[name]["type"] if name in props else "str"

    # NOTE: convert
    if value_type == "size" and direction == "to":
        return globals()[f"{direction}_{value_type}"](value, convert_to_human)

    return globals()[f"{direction}_{value_type}"](value)


@real_memoize
def _zfs_cmd():
    """
    Return the path of the zfs binary if present
    """
    # Get the path to the zfs binary.
    return salt.utils.path.which("zfs")


@real_memoize
def _zpool_cmd():
    """
    Return the path of the zpool binary if present
    """
    # Get the path to the zfs binary.
    return salt.utils.path.which("zpool")


def _command(
    source,
    command,
    flags=None,
    opts=None,
    property_name=None,
    property_value=None,
    filesystem_properties=None,
    pool_properties=None,
    target=None,
):
    """
    Build and properly escape a zfs command

    .. note::

        Input is not considered safe and will be passed through
        to_auto(from_auto('input_here')), you do not need to do so
        your self first.

    """
    # NOTE: start with the zfs binary and command
    cmd = [_zpool_cmd() if source == "zpool" else _zfs_cmd(), command]

    # NOTE: append flags if we have any
    if flags is None:
        flags = []
    for flag in flags:
        cmd.append(flag)

    # NOTE: append options
    #       we pass through 'sorted' to guarantee the same order
    if opts is None:
        opts = {}
    for opt in sorted(opts):
        if not isinstance(opts[opt], list):
            opts[opt] = [opts[opt]]
        for val in opts[opt]:
            cmd.append(opt)
            cmd.append(to_str(val))

    # NOTE: append filesystem properties (really just options with a key/value)
    #       we pass through 'sorted' to guarantee the same order
    if filesystem_properties is None:
        filesystem_properties = {}
    for fsopt in sorted(filesystem_properties):
        cmd.append("-O" if source == "zpool" else "-o")
        cmd.append(
            "{key}={val}".format(
                key=fsopt,
                val=to_auto(
                    fsopt,
                    filesystem_properties[fsopt],
                    source="zfs",
                    convert_to_human=False,
                ),
            )
        )

    # NOTE: append pool properties (really just options with a key/value)
    #       we pass through 'sorted' to guarantee the same order
    if pool_properties is None:
        pool_properties = {}
    for fsopt in sorted(pool_properties):
        cmd.append("-o")
        cmd.append(
            "{key}={val}".format(
                key=fsopt,
                val=to_auto(
                    fsopt,
                    pool_properties[fsopt],
                    source="zpool",
                    convert_to_human=False,
                ),
            )
        )

    # NOTE: append property and value
    #       the set command takes a key=value pair, we need to support this
    if property_name is not None:
        if property_value is not None:
            if not isinstance(property_name, list):
                property_name = [property_name]
            if not isinstance(property_value, list):
                property_value = [property_value]
            for key, val in zip(property_name, property_value):
                cmd.append(
                    "{key}={val}".format(
                        key=key,
                        val=to_auto(key, val, source=source, convert_to_human=False),
                    )
                )
        else:
            cmd.append(property_name)

    # NOTE: append the target(s)
    if target is not None:
        if not isinstance(target, list):
            target = [target]
        for tgt in target:
            # NOTE: skip None list items
            #       we do not want to skip False and 0!
            if tgt is None:
                continue
            cmd.append(to_str(tgt))

    return " ".join(cmd)


def is_supported():
    """
    Check the system for ZFS support
    """
    # Check for supported platforms
    # NOTE: ZFS on Windows is in development
    # NOTE: ZFS on NetBSD is in development
    on_supported_platform = False
    if salt.utils.platform.is_sunos():
        on_supported_platform = True
    elif salt.utils.platform.is_freebsd() and _check_retcode("kldstat -q -m zfs"):
        on_supported_platform = True
    elif salt.utils.platform.is_linux() and os.path.exists("/sys/module/zfs"):
        on_supported_platform = True
    elif salt.utils.platform.is_linux() and salt.utils.path.which("zfs-fuse"):
        on_supported_platform = True
    elif (
        salt.utils.platform.is_darwin()
        and os.path.exists("/Library/Extensions/zfs.kext")
        and os.path.exists("/dev/zfs")
    ):
        on_supported_platform = True

    # Additional check for the zpool command
    return (salt.utils.path.which("zpool") and on_supported_platform) is True


@real_memoize
def has_feature_flags():
    """
    Check if zpool-features is available
    """
    # get man location
    man = salt.utils.path.which("man")
    return _check_retcode(f"{man} zpool-features") if man else False


@real_memoize
def property_data_zpool():
    """
    Return a dict of zpool properties

    .. note::

        Each property will have an entry with the following info:
            - edit : boolean - is this property editable after pool creation
            - type : str - either bool, bool_alt, size, numeric, or string
            - values : str - list of possible values

    .. warning::

        This data is probed from the output of 'zpool get' with some supplemental
        data that is hardcoded. There is no better way to get this information aside
        from reading the code.

    """
    # NOTE: man page also mentions a few short forms
    property_data = _property_parse_cmd(
        _zpool_cmd(),
        {
            "allocated": "alloc",
            "autoexpand": "expand",
            "autoreplace": "replace",
            "listsnapshots": "listsnaps",
            "fragmentation": "frag",
        },
    )

    # NOTE: zpool status/iostat has a few extra fields
    zpool_size_extra = [
        "capacity-alloc",
        "capacity-free",
        "operations-read",
        "operations-write",
        "bandwidth-read",
        "bandwidth-write",
        "read",
        "write",
    ]
    zpool_numeric_extra = [
        "cksum",
        "cap",
    ]

    for prop in zpool_size_extra:
        property_data[prop] = {
            "edit": False,
            "type": "size",
            "values": "<size>",
        }

    for prop in zpool_numeric_extra:
        property_data[prop] = {
            "edit": False,
            "type": "numeric",
            "values": "<count>",
        }

    return property_data


@real_memoize
def property_data_zfs():
    """
    Return a dict of zfs properties

    .. note::

        Each property will have an entry with the following info:
            - edit : boolean - is this property editable after pool creation
            - inherit : boolean - is this property inheritable
            - type : str - either bool, bool_alt, size, numeric, or string
            - values : str - list of possible values

    .. warning::

        This data is probed from the output of 'zfs get' with some supplemental
        data that is hardcoded. There is no better way to get this information aside
        from reading the code.

    """
    return _property_parse_cmd(
        _zfs_cmd(),
        {
            "available": "avail",
            "logicalreferenced": "lrefer.",
            "logicalused": "lused.",
            "referenced": "refer",
            "volblocksize": "volblock",
            "compression": "compress",
            "readonly": "rdonly",
            "recordsize": "recsize",
            "refreservation": "refreserv",
            "reservation": "reserv",
        },
    )


def from_numeric(value):
    """
    Convert zfs numeric to python int
    """
    if value == "none":
        value = None
    elif value:
        value = str_to_num(value)
    return value


def to_numeric(value):
    """
    Convert python int to zfs numeric
    """
    value = from_numeric(value)
    if value is None:
        value = "none"
    return value


def from_bool(value):
    """
    Convert zfs bool to python bool
    """
    if value in ["on", "yes"]:
        value = True
    elif value in ["off", "no"]:
        value = False
    elif value == "none":
        value = None

    return value


def from_bool_alt(value):
    """
    Convert zfs bool_alt to python bool
    """
    return from_bool(value)


def to_bool(value):
    """
    Convert python bool to zfs on/off bool
    """
    value = from_bool(value)
    if isinstance(value, bool):
        value = "on" if value else "off"
    elif value is None:
        value = "none"

    return value


def to_bool_alt(value):
    """
    Convert python to zfs yes/no value
    """
    value = from_bool_alt(value)
    if isinstance(value, bool):
        value = "yes" if value else "no"
    elif value is None:
        value = "none"

    return value


def from_size(value):
    """
    Convert zfs size (human readable) to python int (bytes)
    """
    match_size = re_zfs_size.match(str(value))
    if match_size:
        v_unit = match_size.group(2).upper()[0]
        v_size = float(match_size.group(1))
        v_multiplier = math.pow(1024, zfs_size.index(v_unit) + 1)
        value = v_size * v_multiplier
        if int(value) == value:
            value = int(value)
    elif value is not None:
        value = str(value)

    return from_numeric(value)


def to_size(value, convert_to_human=True):
    """
    Convert python int (bytes) to zfs size

    NOTE: http://src.illumos.org/source/xref/illumos-gate/usr/src/lib/pyzfs/common/util.py#114
    """
    value = from_size(value)
    if value is None:
        value = "none"

    if isinstance(value, Number) and value > 1024 and convert_to_human:
        v_power = int(math.floor(math.log(value, 1024)))
        v_multiplier = math.pow(1024, v_power)

        # NOTE: zfs is a bit odd on how it does the rounding,
        #       see libzfs implementation linked above
        v_size_float = float(value) / v_multiplier
        if v_size_float == int(v_size_float):
            value = "{:.0f}{}".format(
                v_size_float,
                zfs_size[v_power - 1],
            )
        else:
            for v_precision in ["{:.2f}{}", "{:.1f}{}", "{:.0f}{}"]:
                v_size = v_precision.format(
                    v_size_float,
                    zfs_size[v_power - 1],
                )
                if len(v_size) <= 5:
                    value = v_size
                    break

    return value


def from_str(value):
    """
    Decode zfs safe string (used for name, path, ...)
    """
    if value == "none":
        value = None
    if value:
        value = str(value)
        if value.startswith('"') and value.endswith('"'):
            value = value[1:-1]
        value = value.replace('\\"', '"')

    return value


def to_str(value):
    """
    Encode zfs safe string (used for name, path, ...)
    """
    value = from_str(value)

    if value:
        value = value.replace('"', '\\"')
        if " " in value:
            value = '"' + value + '"'
    elif value is None:
        value = "none"

    return value


def from_auto(name, value, source="auto"):
    """
    Convert zfs value to python value
    """
    return _auto("from", name, value, source)


def to_auto(name, value, source="auto", convert_to_human=True):
    """
    Convert python value to zfs value
    """
    return _auto("to", name, value, source, convert_to_human)


def from_auto_dict(values, source="auto"):
    """
    Pass an entire dictionary to from_auto

    .. note::
        The key will be passed as the name

    """
    for name, value in values.items():
        values[name] = from_auto(name, value, source)

    return values


def to_auto_dict(values, source="auto", convert_to_human=True):
    """
    Pass an entire dictionary to to_auto

    .. note::
        The key will be passed as the name
    """
    for name, value in values.items():
        values[name] = to_auto(name, value, source, convert_to_human)

    return values


def is_snapshot(name):
    """
    Check if name is a valid snapshot name
    """
    return from_str(name).count("@") == 1


def is_bookmark(name):
    """
    Check if name is a valid bookmark name
    """
    return from_str(name).count("#") == 1


def is_dataset(name):
    """
    Check if name is a valid filesystem or volume name
    """
    return not is_snapshot(name) and not is_bookmark(name)


def zfs_command(
    command,
    flags=None,
    opts=None,
    property_name=None,
    property_value=None,
    filesystem_properties=None,
    target=None,
):
    """
    Build and properly escape a zfs command

    .. note::

        Input is not considered safe and will be passed through
        to_auto(from_auto('input_here')), you do not need to do so
        your self first.

    """
    return _command(
        "zfs",
        command=command,
        flags=flags,
        opts=opts,
        property_name=property_name,
        property_value=property_value,
        filesystem_properties=filesystem_properties,
        pool_properties=None,
        target=target,
    )


def zpool_command(
    command,
    flags=None,
    opts=None,
    property_name=None,
    property_value=None,
    filesystem_properties=None,
    pool_properties=None,
    target=None,
):
    """
    Build and properly escape a zpool command

    .. note::

        Input is not considered safe and will be passed through
        to_auto(from_auto('input_here')), you do not need to do so
        your self first.

    """
    return _command(
        "zpool",
        command=command,
        flags=flags,
        opts=opts,
        property_name=property_name,
        property_value=property_value,
        filesystem_properties=filesystem_properties,
        pool_properties=pool_properties,
        target=target,
    )


def parse_command_result(res, label=None):
    """
    Parse the result of a zpool/zfs command

    .. note::

        Output on failure is rather predictable.
        - retcode > 0
        - each 'error' is a line on stderr
        - optional 'Usage:' block under those with hits

        We simple check those and return a OrderedDict were
        we set label = True|False and error = error_messages

    """
    ret = OrderedDict()

    if label:
        ret[label] = res["retcode"] == 0

    if res["retcode"] != 0:
        ret["error"] = []
        for error in res["stderr"].splitlines():
            if error.lower().startswith("usage:"):
                break
            if error.lower().startswith("use '-f'"):
                error = error.replace("-f", "force=True")
            if error.lower().startswith("use '-r'"):
                error = error.replace("-r", "recursive=True")
            ret["error"].append(error)

        if ret["error"]:
            ret["error"] = "\n".join(ret["error"])
        else:
            del ret["error"]

    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.boto_dynamodb as boto_dynamodb
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {boto_dynamodb: {}}


def test_present():
    """
    Test to ensure the DynamoDB table exists.
    """
    name = "new_table"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    exists_mock = MagicMock(side_effect=[True, False, False])
    dict_mock = MagicMock(return_value={})
    mock_bool = MagicMock(return_value=True)
    pillar_mock = MagicMock(return_value=[])
    with patch.dict(
        boto_dynamodb.__salt__,
        {
            "boto_dynamodb.exists": exists_mock,
            "boto_dynamodb.describe": dict_mock,
            "config.option": dict_mock,
            "pillar.get": pillar_mock,
            "boto_dynamodb.create_table": mock_bool,
        },
    ):
        comt = (
            "DynamoDB table {0} exists,\n"
            "DynamoDB table {0} throughput matches,\n"
            "All global secondary indexes match,\n".format(name)
        )
        ret.update({"comment": comt})
        assert boto_dynamodb.present(name) == ret

        with patch.dict(boto_dynamodb.__opts__, {"test": True}):
            comt = f"DynamoDB table {name} would be created."
            ret.update({"comment": comt, "result": None})
            assert boto_dynamodb.present(name) == ret

        changes = {
            "new": {
                "global_indexes": None,
                "hash_key": None,
                "hash_key_data_type": None,
                "local_indexes": None,
                "range_key": None,
                "range_key_data_type": None,
                "read_capacity_units": None,
                "table": "new_table",
                "write_capacity_units": None,
            }
        }

        with patch.dict(boto_dynamodb.__opts__, {"test": False}):
            comt = (
                "DynamoDB table {} was successfully created,\n"
                "DynamoDB table new_table throughput matches,\n".format(name)
            )
            ret.update({"comment": comt, "result": True, "changes": changes})
            assert ret == boto_dynamodb.present(name)


def test_absent():
    """
    Test to ensure the DynamoDB table does not exist.
    """
    name = "new_table"

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    mock = MagicMock(side_effect=[False, True, True])
    mock_bool = MagicMock(return_value=True)
    with patch.dict(
        boto_dynamodb.__salt__,
        {"boto_dynamodb.exists": mock, "boto_dynamodb.delete": mock_bool},
    ):
        comt = f"DynamoDB table {name} does not exist"
        ret.update({"comment": comt})
        assert boto_dynamodb.absent(name) == ret

        with patch.dict(boto_dynamodb.__opts__, {"test": True}):
            comt = f"DynamoDB table {name} is set to be deleted"
            ret.update({"comment": comt, "result": None})
            assert boto_dynamodb.absent(name) == ret

        changes = {
            "new": "Table new_table deleted",
            "old": "Table new_table exists",
        }

        with patch.dict(boto_dynamodb.__opts__, {"test": False}):
            comt = f"Deleted DynamoDB table {name}"
            ret.update({"comment": comt, "result": True, "changes": changes})
            assert boto_dynamodb.absent(name) == ret



"""
Network tools to run from the Master
"""

import logging
import socket

import salt.utils.files
import salt.utils.network
import salt.utils.stringutils

log = logging.getLogger(__name__)


def wollist(maclist, bcast="255.255.255.255", destport=9):
    """
    Send a "Magic Packet" to wake up a list of Minions.
    This list must contain one MAC hardware address per line

    CLI Example:

    .. code-block:: bash

        salt-run network.wollist '/path/to/maclist'
        salt-run network.wollist '/path/to/maclist' 255.255.255.255 7
        salt-run network.wollist '/path/to/maclist' 255.255.255.255 7
    """
    ret = []
    try:
        with salt.utils.files.fopen(maclist, "r") as ifile:
            for mac in ifile:
                mac = salt.utils.stringutils.to_unicode(mac).strip()
                wol(mac, bcast, destport)
                print(f"Waking up {mac}")
                ret.append(mac)
    except Exception as err:  # pylint: disable=broad-except
        __jid_event__.fire_event(
            {"error": f"Failed to open the MAC file. Error: {err}"}, "progress"
        )
        return []
    return ret


def wol(mac, bcast="255.255.255.255", destport=9):
    """
    Send a "Magic Packet" to wake up a Minion

    CLI Example:

    .. code-block:: bash

        salt-run network.wol 08-00-27-13-69-77
        salt-run network.wol 080027136977 255.255.255.255 7
        salt-run network.wol 08:00:27:13:69:77 255.255.255.255 7
    """
    dest = salt.utils.network.mac_str_to_bytes(mac)
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    sock.sendto(b"\xff" * 6 + dest * 16, (bcast, int(destport)))
    return True


def wolmatch(tgt, tgt_type="glob", bcast="255.255.255.255", destport=9):
    """
    Send a "Magic Packet" to wake up Minions that are matched in the grains cache

    CLI Example:

    .. code-block:: bash

        salt-run network.wolmatch minion_id
        salt-run network.wolmatch 192.168.0.0/16 tgt_type='ipcidr' bcast=255.255.255.255 destport=7
    """
    ret = []
    minions = __salt__["cache.grains"](tgt, tgt_type)
    for minion in minions:
        for iface, mac in minions[minion]["hwaddr_interfaces"].items():
            if iface == "lo":
                continue
            mac = mac.strip()
            wol(mac, bcast, destport)
            log.info("Waking up %s", mac)
            ret.append(mac)
    return ret



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>

    Test cases for salt.modules.nfs3
"""

import pytest

import salt.modules.nfs3 as nfs3
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {nfs3: {}}


def test_list_exports():
    """
    Test for List configured exports
    """
    with patch("salt.utils.files.fopen", mock_open(read_data="A B1(23")):
        exports = nfs3.list_exports()
        assert exports == {"A": [{"hosts": "B1", "options": ["23"]}]}, exports


def test_del_export():
    """
    Test for Remove an export
    """
    list_exports_mock = MagicMock(
        return_value={"A": [{"hosts": ["B1"], "options": ["23"]}]}
    )
    with patch.object(nfs3, "list_exports", list_exports_mock), patch.object(
        nfs3, "_write_exports", MagicMock(return_value=None)
    ):
        result = nfs3.del_export(path="A")
        assert result == {}, result



"""
Management of NTP servers
=========================

.. versionadded:: 2014.1.0

This state is used to manage NTP servers. Currently only Windows is supported.

.. code-block:: yaml

    win_ntp:
      ntp.managed:
        - servers:
          - pool.ntp.org
          - us.pool.ntp.org
"""

import logging

import salt.utils.platform

log = logging.getLogger(__name__)


def __virtual__():
    """
    This only supports Windows
    """
    if not salt.utils.platform.is_windows():
        return (False, "Only Windows supported")
    return "ntp"


def _check_servers(servers):
    if not isinstance(servers, list):
        return False
    for server in servers:
        if not isinstance(server, str):
            return False
    return True


def _get_servers():
    try:
        return set(__salt__["ntp.get_servers"]())
    except TypeError:
        return {False}


def managed(name, servers=None):
    """
    Manage NTP servers

    servers
        A list of NTP servers
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "NTP servers already configured as specified",
    }

    if not _check_servers(servers):
        ret["result"] = False
        ret["comment"] = "NTP servers must be a list of strings"

    before_servers = _get_servers()
    desired_servers = set(servers)

    if before_servers == desired_servers:
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "NTP servers will be updated to: {}".format(
            ", ".join(sorted(desired_servers))
        )
        return ret

    __salt__["ntp.set_servers"](*desired_servers)
    after_servers = _get_servers()

    if after_servers == desired_servers:
        ret["comment"] = "NTP servers updated"
        ret["changes"] = {"old": sorted(before_servers), "new": sorted(after_servers)}
    else:
        ret["result"] = False
        ret["comment"] = "Failed to update NTP servers"
        if before_servers != after_servers:
            ret["changes"] = {
                "old": sorted(before_servers),
                "new": sorted(after_servers),
            }

    return ret



import pytest

import tests.support.netapi as netapi


@pytest.fixture
def client_config(client_config, netapi_port):
    client_config["rest_tornado"] = {"port": netapi_port}
    client_config["netapi_enable_clients"] = [
        "local",
        "local_async",
        "runner",
        "runner_async",
    ]
    return client_config


@pytest.fixture
def app(app_urls, load_auth, client_config, minion_config):
    return netapi.build_tornado_app(app_urls, load_auth, client_config, minion_config)


@pytest.fixture
def http_server(io_loop, app, netapi_port):
    with netapi.TestsTornadoHttpServer(
        io_loop=io_loop, app=app, port=netapi_port
    ) as server:
        yield server


@pytest.fixture
def http_client(http_server):
    return http_server.client



"""
Display output for minions that did not return
==============================================

This outputter is used to display notices about which minions failed to return
when a salt function is run with ``-v`` or ``--verbose``. It should not be
called directly from the CLI.

Example output::

    virtucentos:
        Minion did not return
"""

import salt.utils.color


class NestDisplay:
    """
    Create generator for nested output
    """

    def __init__(self):
        self.colors = salt.utils.color.get_colors(
            __opts__.get("color"), __opts__.get("color_theme")
        )

    def display(self, ret, indent, prefix, out):
        """
        Recursively iterate down through data structures to determine output
        """
        if isinstance(ret, str):
            lines = ret.split("\n")
            for line in lines:
                out += "{}{}{}{}{}\n".format(
                    self.colors["RED"], " " * indent, prefix, line, self.colors["ENDC"]
                )
        elif isinstance(ret, dict):
            for key in sorted(ret):
                val = ret[key]
                out += "{}{}{}{}{}:\n".format(
                    self.colors["CYAN"], " " * indent, prefix, key, self.colors["ENDC"]
                )
                out = self.display(val, indent + 4, "", out)
        return out


def output(ret, **kwargs):  # pylint: disable=unused-argument
    """
    Display ret data
    """
    nest = NestDisplay()
    return nest.display(ret, 0, "", "")



from textwrap import dedent

import pytest

import salt.utils.path

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture(scope="module")
def cmd(modules):
    return modules.cmd


@pytest.fixture(scope="module")
def account():
    with pytest.helpers.create_account() as _account:
        yield _account


@pytest.fixture(scope="module")
def exitcode_script(state_tree):
    exit_code = 12345
    script_contents = dedent(
        f"""\
        Write-Host "Expected exit code: {exit_code}"
        exit {exit_code}
        """
    )
    with pytest.helpers.temp_file("exit_code.ps1", script_contents, state_tree):
        yield exit_code


@pytest.fixture(scope="module")
def echo_script(state_tree):
    exit_code = 12345
    script_contents = dedent(
        """\
        param (
            [string]$a,
            [string]$b
        )
        Write-Output "a: $a, b: $b"
        """
    )
    with pytest.helpers.temp_file("echo.ps1", script_contents, state_tree):
        yield exit_code


@pytest.fixture(params=["powershell", "pwsh"])
def shell(request):
    """
    This will run the test on PowerShell and PowerShell core (pwsh). If
    PowerShell core is not installed, that test run will be skipped
    """
    if request.param == "pwsh" and salt.utils.path.which("pwsh") is None:
        pytest.skip("Powershell 7 Not Present")
    return request.param


def test_exitcode(cmd, shell, exitcode_script):
    """
    Test receiving an exit code from a PowerShell script
    """
    ret = cmd.script("salt://exit_code.ps1", shell=shell, saltenv="base")
    assert ret["retcode"] == exitcode_script


@pytest.mark.parametrize(
    "args, expected",
    [
        ("foo bar", "a: foo, b: bar"),
        ('foo "bar bar"', "a: foo, b: bar bar"),
        (["foo", "bar"], "a: foo, b: bar"),
        (["foo foo", "bar bar"], "a: foo foo, b: bar bar"),
    ],
)
def test_echo(cmd, shell, echo_script, args, expected):
    """
    Test argument processing with a PowerShell script
    """
    ret = cmd.script("salt://echo.ps1", args=args, shell=shell, saltenv="base")
    assert isinstance(ret["pid"], int)
    assert ret["retcode"] == 0
    assert ret["stderr"] == ""
    assert ret["stdout"] == expected


@pytest.mark.parametrize(
    "args, expected",
    [
        ("foo bar", "a: foo, b: bar"),
        ('foo "bar bar"', "a: foo, b: bar bar"),
        (["foo", "bar"], "a: foo, b: bar"),
        (["foo foo", "bar bar"], "a: foo foo, b: bar bar"),
    ],
)
def test_echo_runas(cmd, shell, account, echo_script, args, expected):
    """
    Test argument processing with a PowerShell script and runas
    """
    ret = cmd.script(
        "salt://echo.ps1",
        args=args,
        shell=shell,
        runas=account.username,
        password=account.password,
        saltenv="base",
    )
    assert isinstance(ret["pid"], int)
    assert ret["retcode"] == 0
    assert ret["stderr"] == ""
    assert ret["stdout"] == expected



"""
Arista pyeapi
=============

.. versionadded:: 2019.2.0

Proxy module for managing Arista switches via the eAPI using the
`pyeapi <http://pyeapi.readthedocs.io/en/master/index.html>`_ library.

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net>
:maturity:   new
:depends:    pyeapi
:platform:   unix

.. note::

    To understand how to correctly enable the eAPI on your switch, please check
    https://eos.arista.com/arista-eapi-101/.


Dependencies
------------

The ``pyeapi`` Proxy module requires pyeapi to be installed:
``pip install pyeapi``.

Pillar
------

The ``pyeapi`` proxy configuration requires the following parameters in order
to connect to the network device:

transport: ``https``
    Specifies the type of connection transport to use. Valid values for the
    connection are ``socket``, ``http_local``, ``http``, and  ``https``.

host: ``localhost``
    The IP address or DNS host name of the connection device.

username: ``admin``
    The username to pass to the device to authenticate the eAPI connection.

password
    The password to pass to the device to authenticate the eAPI connection.

port
    The TCP port of the endpoint for the eAPI connection. If this keyword is
    not specified, the default value is automatically determined by the
    transport type (``80`` for ``http``, or ``443`` for ``https``).

enablepwd
    The enable mode password if required by the destination node.

All the arguments may be optional, depending on your setup.

Proxy Pillar Example
--------------------

.. code-block:: yaml

    proxy:
      proxytype: pyeapi
      host: router1.example.com
      username: example
      password: example
"""

import logging

# Import salt modules
from salt.utils.args import clean_kwargs

try:
    import pyeapi

    HAS_PYEAPI = True
except ImportError:
    HAS_PYEAPI = False


# -----------------------------------------------------------------------------
# proxy properties
# -----------------------------------------------------------------------------

__proxyenabled__ = ["pyeapi"]
# proxy name

# -----------------------------------------------------------------------------
# globals
# -----------------------------------------------------------------------------

__virtualname__ = "pyeapi"
log = logging.getLogger(__name__)
pyeapi_device = {}

# -----------------------------------------------------------------------------
# property functions
# -----------------------------------------------------------------------------


def __virtual__():
    """
    Proxy module available only if pyeapi is installed.
    """
    if not HAS_PYEAPI:
        return (
            False,
            "The pyeapi proxy module requires the pyeapi library to be installed.",
        )
    return __virtualname__


# -----------------------------------------------------------------------------
# proxy functions
# -----------------------------------------------------------------------------


def init(opts):
    """
    Open the connection to the Arista switch over the eAPI.
    """
    proxy_dict = opts.get("proxy", {})
    conn_args = proxy_dict.copy()
    conn_args.pop("proxytype", None)
    opts["multiprocessing"] = conn_args.get("multiprocessing", True)
    # This is not a SSH-based proxy, so it should be safe to enable
    # multiprocessing.
    try:
        conn = pyeapi.client.connect(**conn_args)
        node = pyeapi.client.Node(conn, enablepwd=conn_args.get("enablepwd"))
        pyeapi_device["connection"] = node
        pyeapi_device["initialized"] = True
        pyeapi_device["up"] = True
    except pyeapi.eapilib.ConnectionError as cerr:
        log.error("Unable to connect to %s", conn_args["host"], exc_info=True)
        return False
    return True


def ping():
    """
    Connection open successfully?
    """
    return pyeapi_device.get("up", False)


def initialized():
    """
    Connection finished initializing?
    """
    return pyeapi_device.get("initialized", False)


def shutdown(opts):
    """
    Closes connection with the device.
    """
    log.debug("Shutting down the pyeapi Proxy Minion %s", opts["id"])


# -----------------------------------------------------------------------------
# callable functions
# -----------------------------------------------------------------------------


def conn():
    """
    Return the connection object.
    """
    return pyeapi_device.get("connection")


def call(method, *args, **kwargs):
    """
    Calls an arbitrary pyeapi method.
    """
    kwargs = clean_kwargs(**kwargs)
    return getattr(pyeapi_device["connection"], method)(*args, **kwargs)



"""
These are where the tests go, so that they can be run using both GitPython
and pygit2.

NOTE: The gitfs.update() has to happen AFTER the setUp is called. This is
because running it inside the setUp will spawn a new singleton, which means
that tests which need to mock the __opts__ will be too late; the setUp will
have created a new singleton that will bypass our mocking. To ensure that
our tests are reliable and correct, we want to make sure that each test
uses a new gitfs object, allowing different manipulations of the opts to be
tested.

Therefore, keep the following in mind:

1. Each test needs to call gitfs.update() *after* any patching, and
    *before* calling the function being tested.
2. Do *NOT* move the gitfs.update() into the setUp.

    :codeauthor: Erik Johnson <erik@saltstack.com>
"""

import logging
import os
import pathlib

import pytest

import salt.ext.tornado.ioloop
import salt.fileserver.gitfs as gitfs
import salt.utils.files
import salt.utils.gitfs
import salt.utils.platform
import salt.utils.win_functions
import salt.utils.yaml
from tests.support.helpers import patched_environ
from tests.support.mock import patch

try:
    import pwd  # pylint: disable=unused-import
except ImportError:
    pass


try:
    import git

    # We still need to use GitPython here for temp repo setup, so we do need to
    # actually import it. But we don't need import pygit2 in this module, we
    # can just use the Version instances imported along with
    # salt.utils.gitfs to check if we have a compatible version.
    HAS_GITPYTHON = (
        salt.utils.gitfs.GITPYTHON_VERSION
        and salt.utils.gitfs.GITPYTHON_VERSION >= salt.utils.gitfs.GITPYTHON_MINVER
    )
except (ImportError, AttributeError):
    HAS_GITPYTHON = False

try:
    HAS_PYGIT2 = (
        salt.utils.gitfs.PYGIT2_VERSION
        and salt.utils.gitfs.PYGIT2_VERSION >= salt.utils.gitfs.PYGIT2_MINVER
        and salt.utils.gitfs.LIBGIT2_VERSION
        and salt.utils.gitfs.LIBGIT2_VERSION >= salt.utils.gitfs.LIBGIT2_MINVER
    )
except AttributeError:
    HAS_PYGIT2 = False

log = logging.getLogger(__name__)


@pytest.fixture(scope="module", params=["gitpython", "pygit2"], autouse=True)
def provider(request):
    if not HAS_GITPYTHON:
        pytest.skip(
            "GitPython >= {} required for temp repo setup".format(
                salt.utils.gitfs.GITPYTHON_MINVER
            )
        )
    if request.param == "pygit2":
        if not HAS_PYGIT2:
            pytest.skip(
                "pygit2 >= {} and libgit2 >= {} required".format(
                    salt.utils.gitfs.PYGIT2_MINVER, salt.utils.gitfs.LIBGIT2_MINVER
                )
            )
        if salt.utils.platform.is_windows():
            pytest.skip("Skip Pygit2 on windows, due to pygit2 access error on windows")

    return request.param


@pytest.fixture
def sock_dir(tmp_path):
    dirname = tmp_path / "sock_dir"
    dirname.mkdir(parents=True, exist_ok=True)
    return dirname


@pytest.fixture(scope="module")
def unicode_filename():
    return "питон.txt"


@pytest.fixture(scope="module")
def tag_name():
    return "mytag"


@pytest.fixture(scope="module")
def unicode_dirname():
    return "соль"


@pytest.fixture(scope="module")
def load():
    return {"saltenv": "base"}


@pytest.fixture(autouse=True)
def testfile(tmp_path):
    fp = tmp_path / "testfile"
    fp.write_text("This is a testfile")
    return fp


@pytest.fixture
def repo_dir(tmp_path, unicode_dirname, tag_name, unicode_filename):
    try:
        del salt.utils.gitfs.GitFS.instance_map[
            salt.ext.tornado.ioloop.IOLoop.current()
        ]
    except KeyError:
        pass

    dirname = str(tmp_path / "repo_dir")
    if salt.utils.platform.is_windows():
        dirname = dirname.replace("\\", "/")

    # Populate repo
    root = pathlib.Path(dirname)
    grail_dir = root / "grail"
    grail_dir.mkdir(parents=True, exist_ok=True)
    (grail_dir / "random_file").touch()
    testfile = root / "testfile"
    testfile.write_text("this is a testfile in a git repo\n")
    unicode_file = root / unicode_filename
    unicode_file.write_text("\nThis is a file with a unicode name\n")
    unicode_dir = root / unicode_dirname
    unicode_dir.mkdir(parents=True, exist_ok=True)
    (unicode_dir / "foo.txt").touch()

    # Generate git data
    repo = git.Repo.init(dirname)
    try:
        if salt.utils.platform.is_windows():
            username = salt.utils.win_functions.get_current_user()
        else:
            username = pwd.getpwuid(os.geteuid()).pw_name
    except AttributeError:
        log.error("Unable to get effective username, falling back to 'root'.")
        username = "root"

    with patched_environ(USERNAME=username):
        repo.index.add([x for x in os.listdir(dirname) if x != ".git"])
        repo.index.commit("Test")

        # Add another branch with unicode characters in the name
        repo.create_head(unicode_dirname, "HEAD")

        # Add a tag
        repo.create_tag(tag_name, "HEAD")
        # Older GitPython versions do not have a close method.
        if hasattr(repo, "close"):
            repo.close()

    return dirname


@pytest.fixture
def cache_dir(tmp_path):
    dirname = tmp_path / "cache_dir"
    dirname.mkdir(parents=True, exist_ok=True)
    return dirname


@pytest.fixture
def configure_loader_modules(provider, sock_dir, repo_dir, cache_dir):
    opts = {
        "sock_dir": str(sock_dir),
        "gitfs_remotes": [f"file://{repo_dir}"],
        "cachedir": str(cache_dir),
        "gitfs_root": "",
        "fileserver_backend": ["gitfs"],
        "gitfs_base": "master",
        "gitfs_fallback": "",
        "fileserver_events": True,
        "transport": "zeromq",
        "gitfs_mountpoint": "",
        "gitfs_saltenv": [],
        "gitfs_saltenv_whitelist": [],
        "gitfs_saltenv_blacklist": [],
        "gitfs_user": "",
        "gitfs_password": "",
        "gitfs_insecure_auth": False,
        "gitfs_privkey": "",
        "gitfs_pubkey": "",
        "gitfs_passphrase": "",
        "gitfs_refspecs": [
            "+refs/heads/*:refs/remotes/origin/*",
            "+refs/tags/*:refs/tags/*",
        ],
        "gitfs_ssl_verify": True,
        "gitfs_disable_saltenv_mapping": False,
        "gitfs_ref_types": ["branch", "tag", "sha"],
        "gitfs_update_interval": 60,
        "__role": "master",
        "gitfs_provider": provider,
    }

    return {gitfs: {"__opts__": opts}}


@pytest.fixture(autouse=True)
def skip_on_no_virtual(configure_loader_modules):
    if not gitfs.__virtual__():
        pytest.skip("GitFS could not be loaded. Skipping GitFS tests!")


@pytest.mark.slow_test
def test_file_list(load, unicode_filename, unicode_dirname):
    gitfs.update()
    ret = gitfs.file_list(load)
    assert "testfile" in ret
    assert unicode_filename in ret
    # This function does not use os.sep, the Salt fileserver uses the
    # forward slash, hence it being explicitly used to join here.
    assert "/".join((unicode_dirname, "foo.txt")) in ret


@pytest.mark.slow_test
def test_dir_list(load, unicode_dirname):
    gitfs.update()
    ret = gitfs.dir_list(load)
    assert "grail" in ret
    assert unicode_dirname in ret


def test_find_and_serve_file(repo_dir):
    with patch.dict(gitfs.__opts__, {"file_buffer_size": 262144}):
        gitfs.update()

        # find_file
        ret = gitfs.find_file("testfile")
        assert "testfile" == ret["rel"]

        full_path_to_file = salt.utils.path.join(
            gitfs.__opts__["cachedir"], "gitfs", "refs", "base", "testfile"
        )
        assert full_path_to_file == ret["path"]

        # serve_file
        load = {"saltenv": "base", "path": full_path_to_file, "loc": 0}
        fnd = {"path": full_path_to_file, "rel": "testfile"}
        ret = gitfs.serve_file(load, fnd)

        with salt.utils.files.fopen(
            os.path.join(repo_dir, "testfile"), "r"
        ) as fp_:  # NB: Why not 'rb'?
            data = fp_.read()

        assert ret == {"data": data, "dest": "testfile"}


def test_file_list_fallback(unicode_filename, unicode_dirname):
    with patch.dict(gitfs.__opts__, {"gitfs_fallback": "master"}):
        gitfs.update()
        ret = gitfs.file_list({"saltenv": "notexisting"})
        assert "testfile" in ret
        assert unicode_filename in ret
        # This function does not use os.sep, the Salt fileserver uses the
        # forward slash, hence it being explicitly used to join here.
        assert "/".join((unicode_dirname, "foo.txt")) in ret


def test_dir_list_fallback(unicode_dirname):
    with patch.dict(gitfs.__opts__, {"gitfs_fallback": "master"}):
        gitfs.update()
        ret = gitfs.dir_list({"saltenv": "notexisting"})
        assert "grail" in ret
        assert unicode_dirname in ret


def test_find_and_serve_file_fallback(repo_dir):
    with patch.dict(
        gitfs.__opts__, {"file_buffer_size": 262144, "gitfs_fallback": "master"}
    ):
        gitfs.update()

        # find_file
        ret = gitfs.find_file("testfile", tgt_env="notexisting")
        assert "testfile" == ret["rel"]

        full_path_to_file = salt.utils.path.join(
            gitfs.__opts__["cachedir"], "gitfs", "refs", "notexisting", "testfile"
        )
        assert full_path_to_file == ret["path"]

        # serve_file
        load = {"saltenv": "notexisting", "path": full_path_to_file, "loc": 0}
        fnd = {"path": full_path_to_file, "rel": "testfile"}
        ret = gitfs.serve_file(load, fnd)

        with salt.utils.files.fopen(
            os.path.join(repo_dir, "testfile"), "r"
        ) as fp_:  # NB: Why not 'rb'?
            data = fp_.read()

        assert ret == {"data": data, "dest": "testfile"}


@pytest.mark.slow_test
def test_envs(unicode_dirname, tag_name):
    gitfs.update()
    ret = gitfs.envs(ignore_cache=True)
    assert "base" in ret
    assert unicode_dirname in ret
    assert tag_name in ret


@pytest.mark.slow_test
def test_ref_types_global(unicode_dirname, tag_name):
    """
    Test the global gitfs_ref_types config option
    """
    with patch.dict(gitfs.__opts__, {"gitfs_ref_types": ["branch"]}):
        gitfs.update()
        ret = gitfs.envs(ignore_cache=True)
        # Since we are restricting to branches only, the tag should not
        # appear in the envs list.
        assert "base" in ret
        assert unicode_dirname in ret
        assert tag_name not in ret


@pytest.mark.slow_test
def test_ref_types_per_remote(repo_dir, unicode_dirname, tag_name):
    """
    Test the per_remote ref_types config option, using a different
    ref_types setting than the global test.
    """
    remotes = [{f"file://{repo_dir}": [{"ref_types": ["tag"]}]}]
    with patch.dict(gitfs.__opts__, {"gitfs_remotes": remotes}):
        gitfs.update()
        ret = gitfs.envs(ignore_cache=True)
        # Since we are restricting to tags only, the tag should appear in
        # the envs list, but the branches should not.
        assert "base" not in ret
        assert unicode_dirname not in ret
        assert tag_name in ret


@pytest.mark.slow_test
def test_disable_saltenv_mapping_global_with_mapping_defined_globally():
    """
    Test the global gitfs_disable_saltenv_mapping config option, combined
    with the per-saltenv mapping being defined in the global gitfs_saltenv
    option.
    """
    opts = {
        "gitfs_disable_saltenv_mapping": True,
        "gitfs_saltenv": [{"foo": [{"ref": "somebranch"}]}],
    }
    with patch.dict(gitfs.__opts__, opts):
        gitfs.update()
        ret = gitfs.envs(ignore_cache=True)
        # Since we are restricting to tags only, the tag should appear in
        # the envs list, but the branches should not.
        assert ret == ["base", "foo"]


@pytest.mark.slow_test
def test_saltenv_blacklist(unicode_dirname):
    """
    test saltenv_blacklist
    """
    opts = {"gitfs_saltenv_blacklist": "base"}
    with patch.dict(gitfs.__opts__, opts):
        gitfs.update()
        ret = gitfs.envs(ignore_cache=True)
        assert "base" not in ret
        assert unicode_dirname in ret
        assert "mytag" in ret


@pytest.mark.slow_test
def test_saltenv_whitelist(unicode_dirname):
    """
    test saltenv_whitelist
    """
    opts = {"gitfs_saltenv_whitelist": "base"}
    with patch.dict(gitfs.__opts__, opts):
        gitfs.update()
        ret = gitfs.envs(ignore_cache=True)
        assert "base" in ret
        assert unicode_dirname not in ret
        assert "mytag" not in ret


@pytest.mark.slow_test
def test_env_deprecated_opts(unicode_dirname):
    """
    ensure deprecated options gitfs_env_whitelist
    and gitfs_env_blacklist do not cause gitfs to
    not load.
    """
    opts = {
        "gitfs_env_whitelist": "base",
        "gitfs_env_blacklist": "",
    }
    with patch.dict(gitfs.__opts__, opts):
        gitfs.update()
        ret = gitfs.envs(ignore_cache=True)
        assert "base" in ret
        assert unicode_dirname in ret
        assert "mytag" in ret


@pytest.mark.slow_test
def test_disable_saltenv_mapping_global_with_mapping_defined_per_remote(repo_dir):
    """
    Test the global gitfs_disable_saltenv_mapping config option, combined
    with the per-saltenv mapping being defined in the remote itself via the
    "saltenv" per-remote option.
    """
    opts = {
        "gitfs_disable_saltenv_mapping": True,
        "gitfs_remotes": [
            {f"file://{repo_dir}": [{"saltenv": [{"bar": [{"ref": "somebranch"}]}]}]}
        ],
    }
    with patch.dict(gitfs.__opts__, opts):
        gitfs.update()
        ret = gitfs.envs(ignore_cache=True)
        # Since we are restricting to tags only, the tag should appear in
        # the envs list, but the branches should not.
        assert ret == ["bar", "base"]


@pytest.mark.slow_test
def test_disable_saltenv_mapping_per_remote_with_mapping_defined_globally(repo_dir):
    """
    Test the per-remote disable_saltenv_mapping config option, combined
    with the per-saltenv mapping being defined in the global gitfs_saltenv
    option.
    """
    opts = {
        "gitfs_remotes": [{f"file://{repo_dir}": [{"disable_saltenv_mapping": True}]}],
        "gitfs_saltenv": [{"hello": [{"ref": "somebranch"}]}],
    }

    with patch.dict(gitfs.__opts__, opts):
        gitfs.update()
        ret = gitfs.envs(ignore_cache=True)
        # Since we are restricting to tags only, the tag should appear in
        # the envs list, but the branches should not.
        assert ret == ["base", "hello"]


@pytest.mark.slow_test
def test_disable_saltenv_mapping_per_remote_with_mapping_defined_per_remote(repo_dir):
    """
    Test the per-remote disable_saltenv_mapping config option, combined
    with the per-saltenv mapping being defined in the remote itself via the
    "saltenv" per-remote option.
    """
    opts = {
        "gitfs_remotes": [
            {
                f"file://{repo_dir}": [
                    {"disable_saltenv_mapping": True},
                    {"saltenv": [{"world": [{"ref": "somebranch"}]}]},
                ]
            }
        ]
    }
    with patch.dict(gitfs.__opts__, opts):
        gitfs.update()
        ret = gitfs.envs(ignore_cache=True)
        # Since we are restricting to tags only, the tag should appear in
        # the envs list, but the branches should not.
        assert ret == ["base", "world"]



"""
Test case for the YAML SDB module
"""

import salt.sdb.yaml as sdb
from tests.support.mock import MagicMock, patch


def test_plaintext():
    """
    Retrieve a value from the top level of the dictionary
    """
    plain = {"foo": "bar"}
    with patch("salt.sdb.yaml._get_values", MagicMock(return_value=plain)):
        assert sdb.get("foo") == "bar"


def test_nested():
    """
    Retrieve a value from a nested level of the dictionary
    """
    plain = {"foo": {"bar": "baz"}}
    with patch("salt.sdb.yaml._get_values", MagicMock(return_value=plain)):
        assert sdb.get("foo:bar") == "baz"


def test_encrypted():
    """
    Assume the content is plaintext if GPG is not configured
    """
    plain = {"foo": "bar"}
    with patch("salt.sdb.yaml._decrypt", MagicMock(return_value=plain)):
        with patch("salt.sdb.yaml._get_values", MagicMock(return_value=None)):
            assert sdb.get("foo", profile={"gpg": True}) == "bar"
